# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Ils series API."""

from functools import partial

from flask import current_app
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.errors import RecordHasReferencesError
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.records_relations.api import IlsRecordWithRelations
from invenio_app_ils.relations.api import (
    MULTIPART_MONOGRAPH_RELATION,
    SERIAL_RELATION,
    ParentChildRelation,
)

SERIES_PID_TYPE = "serid"
SERIES_PID_MINTER = "serid"
SERIES_PID_FETCHER = "serid"

SeriesIdProvider = type(
    "SeriesIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=SERIES_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
series_pid_minter = partial(pid_minter, provider_cls=SeriesIdProvider)
series_pid_fetcher = partial(pid_fetcher, provider_cls=SeriesIdProvider)


class Series(IlsRecordWithRelations):
    """Series record class."""

    SERIES_TYPES = ["SERIAL", "PERIODICAL"]

    _pid_type = SERIES_PID_TYPE
    _schema = "series/series-v2.0.0.json"
    _relations_path = "{scheme}://{host}/api/resolver/series/{series_pid}/relations"

    MODE_OF_ISSUANCE = [
        "MULTIPART_MONOGRAPH",
        "SERIAL",
    ]

    @classmethod
    def build_resolver_fields(cls, data):
        """Build all resolver fields."""
        data.setdefault("relations", {})
        data["relations"] = {
            "$ref": cls._relations_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                series_pid=data["pid"],
            )
        }

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Series record."""
        cls.build_resolver_fields(data)
        return super().create(data, id_=id_, **kwargs)

    def update(self, *args, **kwargs):
        """Update Series record."""
        super().update(*args, **kwargs)
        self.build_resolver_fields(self)

    def delete(self, **kwargs):
        """Delete record with relations."""
        related_refs = set()
        pcr_multipart = ParentChildRelation(MULTIPART_MONOGRAPH_RELATION)
        pcr_serial = ParentChildRelation(SERIAL_RELATION)
        related_multipart_volumes = pcr_multipart.get_children_of(self.pid)
        related_serial_volumes = pcr_serial.get_children_of(self.pid)
        for child_pid in related_serial_volumes + related_multipart_volumes:
            related_refs.add(
                "{pid_value}:{pid_type}".format(
                    pid_value=child_pid.pid_value, pid_type=child_pid.pid_type
                )
            )

        if related_refs:
            raise RecordHasReferencesError(
                record_type=self.__class__.__name__,
                record_id=self["pid"],
                ref_type="volumes",
                ref_ids=sorted(ref for ref in related_refs),
            )
        super(Series, self).delete(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test email notifications on transitions."""

import random
from datetime import timedelta

import arrow
from flask import current_app, url_for
from invenio_circulation.api import Loan
from invenio_circulation.proxies import current_circulation
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from invenio_app_ils.circulation.notifications.tasks import (
    send_expiring_loans_notification_reminder,
    send_overdue_loans_notification_reminder,
)
from invenio_app_ils.documents.api import Document
from invenio_app_ils.patrons.api import Patron
from tests.helpers import user_login, user_logout


def test_notifs_on_loan_checkout(client, app_with_notifs, users, testdata, loan_params):
    """Test that an email is sent on loan checkout."""
    loan_data = testdata["loans"][1]
    loan = Loan.get_record_by_pid(loan_data["pid"])
    with app_with_notifs.extensions["mail"].record_messages() as outbox:
        user_login(client, "admin", users)

        assert len(outbox) == 0
        current_circulation.circulation.trigger(
            loan, **dict(loan_params, trigger="checkout")
        )
        assert len(outbox) == 1
        msg = outbox[0]

    doc = Document.get_record_by_pid(loan_data["document_pid"])
    expected_subject = 'InvenioILS: loan started for "{0}"'.format(doc["title"])
    assert msg.subject == expected_subject

    edition_year = " ({edition} - {year})".format(
        edition=doc["edition"], year=doc["publication_year"]
    )
    full_title = "{title}, {author}{edition_year}".format(
        title=doc["title"],
        author=doc["authors"][0]["full_name"],
        edition_year=edition_year,
    )

    literature_url = "{host}{path}".format(
        host=current_app.config["SPA_HOST"],
        path=current_app.config["SPA_PATHS"]["literature"] % {"pid": doc["pid"]},
    )
    profile_url = "{host}{path}".format(
        host=current_app.config["SPA_HOST"],
        path=current_app.config["SPA_PATHS"]["profile"],
    )
    expected_body_plain = """Dear Patron One,

your loan for "{doc_full_title}" <{literature_url}> has started.

The due date is {loan_end_date}.

You can see your ongoing and past loans in your profile page <{profile_url}>.

Kind regards,
InvenioILS""".format(
        doc_full_title=full_title,
        literature_url=literature_url,
        loan_end_date=loan_data["end_date"],
        profile_url=profile_url,
    )
    assert msg.body == expected_body_plain


def test_notification_on_overdue_permissions(client, testdata, json_headers, users):
    """Test that only the backoffice can send a reminder."""
    pid = testdata["loans"][0]["pid"]
    url = url_for("invenio_app_ils_circulation.loanid_notification", pid_value=pid)
    tests = [("patron1", 403), ("anonymous", 401)]
    for username, expected_status in tests:
        user_login(client, username, users)
        res = client.post(url, headers=json_headers)
        assert res.status_code == expected_status


def test_notification_on_overdue_loans(
    app_with_notifs, db, users, testdata, mocker, client, json_headers
):
    """Test that an email is sent for a loan that is overdue."""
    mocker.patch(
        "invenio_app_ils.patrons.api.Patron.get_patron",
        return_value=Patron(users["patron1"].id),
    )

    def prepare_data():
        """Prepare data."""
        days = current_app.config[
            "ILS_CIRCULATION_NOTIFICATION_OVERDUE_REMINDER_INTERVAL"
        ]
        loans = testdata["loans"]

        recs = []
        now = arrow.utcnow()

        def new_end_date(loan, date):
            loan["end_date"] = date.date().isoformat()
            loan["state"] = "ITEM_ON_LOAN"
            loan.commit()
            recs.append(loan)

        # overdue loans
        date = now - timedelta(days=days)
        new_end_date(loans[0], date)

        date = now - timedelta(days=days * 2)
        new_end_date(loans[1], date)

        # not overdue
        date = now - timedelta(days=-1)
        new_end_date(loans[2], date)

        # not overdue or overdue but not to be notified
        remaining_not_overdue = loans[3:]
        for loan in remaining_not_overdue:
            days = random.choice([-1, 0, 1])
            date = now - timedelta(days=days)
            new_end_date(loan, date)
        db.session.commit()

        indexer = RecordIndexer()
        for rec in recs:
            indexer.index(rec)

        current_search.flush_and_refresh(index="*")

    user_login(client, "librarian", users)

    # test individual overdue loan
    prepare_data()
    loans = testdata["loans"]

    notification_url = url_for(
        "invenio_app_ils_circulation.loanid_notification",
        pid_value=loans[0]["pid"],
    )

    res = client.post(notification_url, headers=json_headers)
    assert res.status_code == 202

    # test individual not overdue loan
    notification_url = url_for(
        "invenio_app_ils_circulation.loanid_notification",
        pid_value=loans[2]["pid"],
    )

    res = client.post(notification_url, headers=json_headers)
    assert res.status_code == 400

    user_logout(client)

    # test all loans
    with app_with_notifs.extensions["mail"].record_messages() as outbox:
        assert len(outbox) == 0
        send_overdue_loans_notification_reminder.apply_async()
        assert len(outbox) == 2


def test_notification_on_expiring_loans(app_with_notifs, db, users, testdata, mocker):
    """Test that an email is sent for a loan that is about to expire."""
    mocker.patch(
        "invenio_app_ils.patrons.api.Patron.get_patron",
        return_value=Patron(users["patron1"].id),
    )

    def prepare_data():
        """Prepare data."""
        max_days = current_app.config["ILS_CIRCULATION_LOAN_WILL_EXPIRE_DAYS"]
        loans = testdata["loans"]

        recs = []
        now = arrow.utcnow()

        def new_end_date(loan, date):
            loan["end_date"] = date.date().isoformat()
            loan["state"] = "ITEM_ON_LOAN"
            loan.commit()
            recs.append(loan)

        # expiring loans
        date = now + timedelta(days=max_days)
        new_end_date(loans[0], date)
        new_end_date(loans[1], date)
        new_end_date(loans[2], date)

        # not expiring
        remaining_not_overdue = loans[3:]
        for loan in remaining_not_overdue:
            days = random.choice([-2, -1, 1, max_days + 1])
            date = now + timedelta(days=days)
            new_end_date(loan, date)
        db.session.commit()

        indexer = RecordIndexer()
        for rec in recs:
            indexer.index(rec)

        current_search.flush_and_refresh(index="*")

    prepare_data()

    with app_with_notifs.extensions["mail"].record_messages() as outbox:
        assert len(outbox) == 0
        send_expiring_loans_notification_reminder.apply_async()
        assert len(outbox) == 3



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Items resolvers."""

from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE

from invenio_app_ils.acquisition.api import ORDER_PID_TYPE
from invenio_app_ils.document_requests.api import DOCUMENT_REQUEST_PID_TYPE
from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE, Document
from invenio_app_ils.documents.indexer import DocumentIndexer
from invenio_app_ils.eitems.api import EITEM_PID_TYPE, EItem
from invenio_app_ils.eitems.indexer import EItemIndexer
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.internal_locations.indexer import InternalLocationIndexer
from invenio_app_ils.items.api import ITEM_PID_TYPE, Item
from invenio_app_ils.items.indexer import ItemIndexer
from invenio_app_ils.locations.api import LOCATION_PID_TYPE, Location
from invenio_app_ils.locations.indexer import LocationIndexer
from invenio_app_ils.patrons.api import PATRON_PID_TYPE, Patron
from invenio_app_ils.patrons.indexer import PatronIndexer

from invenio_app_ils.internal_locations.api import (  # isort:skip
    INTERNAL_LOCATION_PID_TYPE,
    InternalLocation,
)


def test_jsonresolvers(testdata, mocker):
    """Test that all records referencing a changed record are re-indexed."""

    def _get_mock():
        return mocker.patch("invenio_app_ils.indexer.ReferencedRecordsIndexer.index")

    def _assert_origin(indexer, pid_type, pid_value):
        """Assert that origin is called and return all referenced."""
        assert indexer.called
        all_calls = indexer.call_args_list
        first_call = all_calls[0]
        indexed, referenced = first_call[0]
        assert indexed["pid_type"] == pid_type
        assert indexed["record"]["pid"] == pid_value

        # call[0] is the list of arguments passed to the function
        # call[0][0] is the origin record, call[0][1] is the referenced record
        all_referenced = []
        for call in all_calls:
            origin, referenced = call[0]
            all_referenced += referenced
        return all_referenced

    def _assert_contains(referenced, pid_type):
        """Assert that given PID is contained in one of the referenced."""
        found = False
        for r in referenced:
            if r["pid_type"] == pid_type:
                found = True
                break
        assert found

    def test_on_document_update():
        """Test document resolvers."""
        indexer = _get_mock()

        pid = "docid-2"
        document = Document.get_record_by_pid(pid)
        DocumentIndexer().index(document)

        referenced = _assert_origin(indexer, DOCUMENT_PID_TYPE, pid)

        # should re-index document request
        n_doc_req = 1  # from test data
        _assert_contains(referenced, DOCUMENT_REQUEST_PID_TYPE)

        # should re-index eitem
        n_eitems = 1  # from test data
        _assert_contains(referenced, EITEM_PID_TYPE)

        # should re-index item
        n_items = 1  # from test data
        _assert_contains(referenced, ITEM_PID_TYPE)

        # should re-index acq
        n_acq = 1  # from test data
        _assert_contains(referenced, ORDER_PID_TYPE)

        # should re-index ill
        n_ill = 1  # from test data
        _assert_contains(referenced, BORROWING_REQUEST_PID_TYPE)

        expected_total = n_doc_req + n_eitems + n_items + n_acq + n_ill
        assert len(referenced) == expected_total

    def test_on_eitem_update():
        """Test eitem resolvers."""
        indexer = _get_mock()

        pid = "eitemid-1"
        eitem = EItem.get_record_by_pid(pid)
        EItemIndexer().index(eitem)

        referenced = _assert_origin(indexer, EITEM_PID_TYPE, pid)

        # should re-index documents
        n_documents = 1  # from test data
        _assert_contains(referenced, DOCUMENT_PID_TYPE)

        assert len(referenced) == n_documents

    def test_on_internal_location_update():
        """Test internal location resolvers."""
        indexer = _get_mock()

        pid = "ilocid-2"
        intloc = InternalLocation.get_record_by_pid(pid)
        InternalLocationIndexer().index(intloc)

        referenced = _assert_origin(indexer, INTERNAL_LOCATION_PID_TYPE, pid)

        # should re-index items
        n_items = 4  # from test data
        _assert_contains(referenced, ITEM_PID_TYPE)

        assert len(referenced) == n_items

    def test_on_item_update():
        """Test item resolvers."""
        indexer = _get_mock()

        # item on loan
        pid = "itemid-56"
        item = Item.get_record_by_pid(pid)
        ItemIndexer().index(item)

        referenced = _assert_origin(indexer, ITEM_PID_TYPE, pid)

        # should re-index loans
        n_loans = 2  # from test data, including pending
        _assert_contains(referenced, CIRCULATION_LOAN_PID_TYPE)

        # should re-index document
        n_documents = 1  # from test data
        _assert_contains(referenced, DOCUMENT_PID_TYPE)

        assert len(referenced) == n_loans + n_documents

    def test_on_location_update():
        """Test location resolvers."""
        indexer = _get_mock()

        pid = "locid-2"
        loc = Location.get_record_by_pid(pid)
        LocationIndexer().index(loc)

        referenced = _assert_origin(indexer, LOCATION_PID_TYPE, pid)

        # should re-index internal locations
        n_intlocs = 1  # from test data
        _assert_contains(referenced, INTERNAL_LOCATION_PID_TYPE)

        # should re-index internal items
        n_items = 1  # from test data
        _assert_contains(referenced, ITEM_PID_TYPE)

        assert len(referenced) == n_intlocs + n_items

    def test_on_patron_update():
        """Test patron resolvers."""
        indexer = _get_mock()

        pid = "2"
        patron = Patron.get_patron(pid)
        PatronIndexer().index(patron)

        referenced = _assert_origin(indexer, PATRON_PID_TYPE, pid)

        # should re-index loans
        n_loans = 3  # from test data
        _assert_contains(referenced, CIRCULATION_LOAN_PID_TYPE)

        # should re-index document request
        n_doc_req = 1  # from test data
        _assert_contains(referenced, DOCUMENT_REQUEST_PID_TYPE)

        # should re-index acq
        n_acq = 1  # from test data
        _assert_contains(referenced, ORDER_PID_TYPE)

        # should re-index ill
        n_ill = 3  # from test data
        _assert_contains(referenced, BORROWING_REQUEST_PID_TYPE)

        expected_total = n_loans + n_doc_req + n_acq + n_ill
        assert len(referenced) == expected_total

    test_on_document_update()
    test_on_eitem_update()
    test_on_internal_location_update()
    test_on_item_update()
    test_on_location_update()
    test_on_patron_update()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Facets tests."""

from datetime import timedelta

import arrow
import pytest
from flask import current_app
from invenio_search.engine import dsl

from invenio_app_ils.facets import (
    date_range_filter,
    default_value_when_missing_filter,
    keyed_range_filter,
    overdue_loans_filter,
)


def test_keyed_range_filter():
    """Test range filter."""
    range_query = {"None": {"lt": 1}, "1+": {"gte": 1}}
    rfilter = keyed_range_filter("field", range_query)

    assert rfilter(["None"]) == dsl.query.Range(field={"lt": 1})
    assert rfilter(["1+"]) == dsl.query.Range(field={"gte": 1})
    assert rfilter(["None", "1+"]) == dsl.query.Range(field={"gte": 1, "lt": 1})


def test_current_ranged_loans_filter(app):
    """Test ranged current loans filter."""
    with app.app_context():
        rfilter = overdue_loans_filter("field")

        current_loans_query = dsl.query.Terms(
            state=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
        )

        overdue = rfilter(["Overdue"])
        field = {"lt": str(arrow.utcnow().date())}
        assert overdue == dsl.query.Bool(
            must=[dsl.query.Range(field=field), current_loans_query]
        )

        upcoming = rfilter(["Upcoming return"])
        field = {
            "gte": str(arrow.utcnow().date()),
            "lte": str((arrow.utcnow() + timedelta(days=7)).date()),
        }
        assert upcoming == dsl.query.Bool(
            must=[dsl.query.Range(field=field), current_loans_query]
        )


def test_default_value_when_missing_filter(app):
    """Test custom exists filter."""
    with app.app_context():
        rfilter = default_value_when_missing_filter("field", "missing val")

        assert rfilter("test") == dsl.query.Terms(field="test")
        assert rfilter("missing val") == dsl.query.Bool(
            **{"must_not": {"exists": {"field": "field"}}}
        )


def test_date_range_filter(app):
    """Test date range filter date validation and query."""

    tests = ["", "a string", "2020-02-02"]

    for input_date in tests:
        from_filter = date_range_filter("field", "gte")
        to_filter = date_range_filter("field", "lte")

        try:
            assert from_filter([input_date]) == dsl.RangeField(
                field={"gte": input_date}
            )
            assert to_filter([input_date]) == dsl.RangeField(field={"lte": input_date})
        except (ValueError, AssertionError):
            with pytest.raises(ValueError):
                from_filter([input_date])
                to_filter([input_date])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-19 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test record delete."""

from flask import url_for

from tests.helpers import user_login


def test_delete_location(client, users, json_headers, testdata):
    """Test DELETE existing location."""
    user_login(client, "admin", users)

    location_pid = "locid-1"
    url = url_for("invenio_records_rest.locid_item", pid_value=location_pid)
    res = client.delete(url, headers=json_headers)
    assert res.status_code == 400

    location_pid = "locid-3"
    url = url_for("invenio_records_rest.locid_item", pid_value=location_pid)
    res = client.delete(url, headers=json_headers)
    assert res.status_code == 204


def test_delete_internal_location(client, users, json_headers, testdata):
    """Test DELETE existing internal_location."""
    user_login(client, "admin", users)

    internal_location_pid = "ilocid-1"
    url = url_for("invenio_records_rest.ilocid_item", pid_value=internal_location_pid)
    res = client.delete(url, headers=json_headers)
    assert res.status_code == 400

    internal_location_pid = "ilocid-3"
    url = url_for("invenio_records_rest.ilocid_item", pid_value=internal_location_pid)
    res = client.delete(url, headers=json_headers)
    assert res.status_code == 204



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Identifiers schema for marshmallow loader."""

from marshmallow import EXCLUDE, Schema, fields


class IdentifierSchema(Schema):
    """Identifier schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    material = fields.Str()
    scheme = fields.Str(required=True)
    value = fields.Str(required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test record permissions."""

import uuid

import pytest
from flask_principal import RoleNeed, identity_loaded
from invenio_access.models import ActionRoles
from invenio_accounts.models import Role
from invenio_records.api import Record
from invenio_search.api import RecordsSearch

from invenio_app_ils.errors import UnauthorizedSearchError
from invenio_app_ils.records.permissions import RecordPermission, create_records_action
from invenio_app_ils.search_permissions import _filter_by_patron
from tests.helpers import user_login


@pytest.mark.parametrize(
    "patron_pid,qs,should_raise",
    [
        ("1", None, False),
        ("1", "", False),
        ("1", "pid:1234", False),
        ("1", "patron_pid:2", True),
        ("1", "patron_pid: 2", True),
        ("1", "patron_pid: '2'", False),
        ("1", "pid:1234 AND patron_pid:2", True),
    ],
)
def test_filter_by_patron(app, patron_pid, qs, should_raise):
    """Test the function filter_by_patron."""
    search = RecordsSearch()
    if should_raise:
        with pytest.raises(UnauthorizedSearchError):
            _filter_by_patron(patron_pid, search, qs)
    else:
        _search, _qs = _filter_by_patron(patron_pid, search, qs)
        term = _search.to_dict()["query"]["bool"]["filter"][0]["term"]
        assert term == {"patron_pid": patron_pid}


def test_record_generic_access(client, db, users, with_access):
    """Test access control for records."""

    tests = [
        ({"foo": "bar"}, "read", True),
        ({"foo": "bar"}, "update", False),
        ({"_access": {"read": [1]}}, "read", True),
        ({"_access": {"read": [2]}}, "read", False),
        ({"_access": {"read": ["records-readers"]}}, "read", True),
        # permission for specific user to create
        ({"_access": {"update": [1]}}, "update", True),
        # checks if the access works for different actions
        ({"_access": {"update": [1]}}, "create", False),
        ({"_access": {"delete": [1]}}, "update", False),
        # delete access for user and librarian
        ({"_access": {"delete": [1, "librarian"]}}, "delete", True),
    ]

    @identity_loaded.connect
    def add_roles_to_identity(sender, identity):
        """Provide additional role to the user."""
        roles = [RoleNeed("records-readers")]
        identity.provides |= set(roles)

    def login_and_test(username):
        user = user_login(client, username, users)
        # Create record
        id = uuid.uuid4()
        record = Record.create(access, id_=id)
        factory = RecordPermission(record, action)
        if user.has_role("admin"):
            # super user can do EVERYTHING
            assert factory.can()
        elif user.has_role("librarian") and action != "delete":
            # librarian should be able to update, create, and read everything
            assert factory.can()
        else:
            assert factory.can() if is_allowed else not factory.can()

    for access, action, is_allowed in tests:
        # Test standard user
        login_and_test("patron1")
        # Test librarian access
        login_and_test("librarian")
        # Test superuser access
        login_and_test("admin")


@pytest.fixture()
def with_role_creator(db):
    """ "Create a new role and assign action."""
    role = Role(name="records-creators")
    db.session.add(role)
    db.session.commit()
    # assign role to the action "create-records"
    ar = ActionRoles.allow(create_records_action, role_id=role.id)
    db.session.add(ar)
    db.session.commit()


def test_record_patron_create(client, db, users, with_role_creator):
    """Test patron create."""

    tests = [
        ({"foo": "bar"}, "create", True),
        ({"foo": "bar"}, "update", False),
        ({"foo": "bar"}, "delete", False),
    ]

    @identity_loaded.connect
    def add_roles_to_identity(sender, identity):
        """Provide additional role to the user."""
        roles = [RoleNeed("records-creators")]
        identity.provides |= set(roles)

    for access, action, is_allowed in tests:
        # create role to be able to create records
        user_login(client, "patron1", users)

        id = uuid.uuid4()
        record = Record.create(access, id_=id)
        factory = RecordPermission(record, action)

        assert factory.can() if is_allowed else not factory.can()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests documents REST APIs permissions."""

import json

from flask import url_for
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from invenio_app_ils.documents.api import Document
from tests.helpers import user_login


def test_open_access_permissions(client, json_headers, testdata, users):
    """Test GET open/close access documents."""
    # set the documents to have read access only by patron2. `_access` should
    # be totally ignored.
    indexer = RecordIndexer()
    doc1 = Document.get_record_by_pid("docid-open-access")
    doc2 = Document.get_record_by_pid("docid-closed-access")
    for doc in [doc1, doc2]:
        doc.update(dict(_access=dict(read=["patron2"])))
        doc.commit()
        db.session.commit()
        indexer.index(doc)
    current_search.flush_and_refresh(index="documents")

    test_data = [
        ("anonymous", "docid-open-access", 200, 1),
        ("patron1", "docid-open-access", 200, 1),
        ("patron2", "docid-open-access", 200, 1),
        ("librarian", "docid-open-access", 200, 1),
        ("admin", "docid-open-access", 200, 1),
        ("anonymous", "docid-closed-access", 401, 0),
        ("patron1", "docid-closed-access", 403, 0),
        ("patron2", "docid-closed-access", 403, 0),
        ("librarian", "docid-closed-access", 200, 1),
        ("admin", "docid-closed-access", 200, 1),
    ]
    for user, pid, status_code, n_hits in test_data:
        # item endpoint
        user_login(client, user, users)
        url = url_for("invenio_records_rest.docid_item", pid_value=pid)
        res = client.get(url, headers=json_headers)
        assert res.status_code == status_code

        # list endpoint
        user_login(client, user, users)
        url = url_for("invenio_records_rest.docid_list", q="pid:{}".format(pid))
        res = client.get(url, headers=json_headers)
        hits = json.loads(res.data.decode("utf-8"))
        assert hits["hits"]["total"] == n_hits


def test_access_permissions(client, json_headers, testdata, users, with_access):
    """Test GET documents with `_access` ignoring `restricted`."""
    # set the documents to have read access only by patron2. `_access` should
    # be taken into account and take precedence over `restricted`.
    indexer = RecordIndexer()
    doc1 = Document.get_record_by_pid("docid-open-access")
    doc2 = Document.get_record_by_pid("docid-closed-access")
    for doc in [doc1, doc2]:
        doc.update(dict(_access=dict(read=[users["patron2"].id])))
        doc.commit()
        db.session.commit()
        indexer.index(doc)
    current_search.flush_and_refresh(index="documents")

    test_data = [
        ("anonymous", "docid-open-access", 401, 0),
        ("patron1", "docid-open-access", 403, 0),
        ("patron2", "docid-open-access", 200, 1),  # should have access
        ("librarian", "docid-open-access", 200, 1),
        ("admin", "docid-open-access", 200, 1),
        ("anonymous", "docid-closed-access", 401, 0),
        ("patron1", "docid-closed-access", 403, 0),
        ("patron2", "docid-closed-access", 200, 1),  # should have access
        ("librarian", "docid-closed-access", 200, 1),
        ("admin", "docid-closed-access", 200, 1),
    ]
    for user, pid, status_code, n_hits in test_data:
        # item endpoint
        user_login(client, user, users)
        url = url_for("invenio_records_rest.docid_item", pid_value=pid)
        res = client.get(url, headers=json_headers)
        assert res.status_code == status_code

        # list endpoint
        user_login(client, user, users)
        url = url_for("invenio_records_rest.docid_list", q="pid:{}".format(pid))
        res = client.get(url, headers=json_headers)
        hits = json.loads(res.data.decode("utf-8"))
        assert hits["hits"]["total"] == n_hits



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""DocumentRequest schema for marshmallow loader."""

from flask import g
from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from invenio_records_rest.schemas.fields.sanitizedhtml import SanitizedHTML
from marshmallow import EXCLUDE, Schema, ValidationError, fields, validate

from invenio_app_ils.document_requests.api import DocumentRequest
from invenio_app_ils.permissions import backoffice_permission


def validate_patron(patron_pid):
    """Validate patron PID."""
    if not backoffice_permission().allows(g.identity):
        if patron_pid != str(g.identity.id):
            raise ValidationError(
                "The authenticated user is not authorized to create or update "
                "a document request for another patron."
            )


class PhysicalItemProviderSchema(Schema):
    """Physical Item Provider schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    pid = fields.Str(required=True)
    pid_type = fields.Str(required=True)


class DocumentRequestSchemaV1(RecordMetadataSchemaJSONV1):
    """Document Request schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    authors = SanitizedHTML()
    document_pid = SanitizedHTML()
    edition = SanitizedHTML()
    internal_note = SanitizedHTML()
    isbn = SanitizedHTML()
    issn = SanitizedHTML()
    issue = SanitizedHTML()
    journal_title = SanitizedHTML()
    legacy_id = fields.Str()
    medium = fields.Str(required=True)
    note = SanitizedHTML()
    page = SanitizedHTML()
    patron_pid = fields.Str(required=True, validate=validate_patron)
    payment_info = SanitizedHTML()
    payment_method = SanitizedHTML()
    physical_item_provider = fields.Nested(PhysicalItemProviderSchema)
    publication_year = fields.Int()
    request_type = fields.Str(required=True)
    decline_reason = fields.Str(
        # add empty string in case the value must be reset
        validate=validate.OneOf(DocumentRequest.DECLINE_TYPES + [""])
    )
    standard_number = SanitizedHTML()
    state = fields.Str(validate=validate.OneOf(DocumentRequest.STATES))
    title = SanitizedHTML(required=True)
    volume = SanitizedHTML()
    publisher = SanitizedHTML()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Mail backend."""
from celery import shared_task
from flask import current_app
from flask_mail import Message
from invenio_mail.tasks import send_email


@shared_task
def send(patrons, msg, **kwargs):
    """Send an email async with Invenio-Mail and log success / errors.

    :param patrons: List of Patron objects.
    :param msg: A NotificationMsg object.
    :param kwargs: extra arguments.
    """
    kwargs["recipients"] = get_recipients(patrons)
    kwargs["subject"] = msg["title"]
    kwargs["body"] = msg["body_plain"]
    kwargs["html"] = msg["body_html"]
    kwargs.setdefault("sender", current_app.config["MAIL_NOTIFY_SENDER"])
    kwargs.setdefault("cc", current_app.config["MAIL_NOTIFY_CC"])
    kwargs.setdefault("bcc", current_app.config["MAIL_NOTIFY_BCC"])
    message = Message(**kwargs)
    full_data = message.__dict__

    send_email(full_data)


def get_recipients(patrons):
    """Return test recipients only if activated."""
    if current_app.config["ILS_MAIL_ENABLE_TEST_RECIPIENTS"]:
        return current_app.config["ILS_MAIL_NOTIFY_TEST_RECIPIENTS"]
    return [patron["email"] for patron in patrons]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications backends."""

from .mail import send as send_email


def notifications_backend_builder(**kwargs):
    """Factory builder to return a list of backend to use to send notif."""
    return [send_email]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS loaders."""

from invenio_records_rest.loaders import marshmallow_loader
from invenio_records_rest.loaders.marshmallow import MarshmallowErrors


def ils_marshmallow_loader(schema_class):
    """Marshmallow loader for JSON requests."""

    def json_loader():
        try:
            return marshmallow_loader(schema_class)()
        except MarshmallowErrors as me:
            for error in me.errors:
                parent_path = [str(x) for x in error["parents"]]
                error["field"] = ".".join([*parent_path, error["field"]])
            raise me

    return json_loader



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS records permissions."""

from flask import current_app
from flask_principal import ActionNeed, RoleNeed, UserNeed
from invenio_access import Permission, any_user
from six import string_types

from invenio_app_ils.permissions import backoffice_access_action

create_records_action = ActionNeed("create-records")


def record_create_permission_factory(record=None):
    """Record create permission factory."""
    return RecordPermission(record=record, action="create")


def record_update_permission_factory(record=None):
    """Record update permission factory."""
    return RecordPermission(record=record, action="update")


def record_delete_permission_factory(record=None):
    """Record delete permission factory."""
    return RecordPermission(record=record, action="delete")


def record_read_permission_factory(record=None):
    """Record read permission factory."""
    return RecordPermission(record=record, action="read")


class RecordPermission(Permission):
    """Record permission.

    - Create action given to librarian, admin and specified users.
    - Read access given to everyone with possibility to hide.
    - Delete access to admin and specified users.
    """

    def __init__(self, record, action):
        """Constructor."""
        self.record = record
        self.current_action = action
        record_needs = self.collect_needs()
        super().__init__(*record_needs)

    def collect_needs(self):
        """Collect permission policy per action."""
        if self.current_action == "read":
            return self.read_permissions()
        elif self.current_action == "create":
            return [create_records_action, backoffice_access_action]
        elif self.current_action == "update":
            return self.record_needs() + [backoffice_access_action]
        else:
            return self.record_needs()

    def read_permissions(self):
        """Define read permission policy per record."""
        if self.is_public():
            return [any_user]
        else:
            return self.record_needs() + [backoffice_access_action]

    def record_explicit_restrictions(self):
        """Return the list of user ids/roles allowed for the given action."""
        if current_app.config.get("ILS_RECORDS_EXPLICIT_PERMISSIONS_ENABLED"):
            return self.record.get("_access", {}).get(self.current_action, [])
        return []

    def record_needs(self):
        """Create needs of the record."""
        needs = []
        for access_entity in self.record_explicit_restrictions():
            try:
                if isinstance(access_entity, string_types):
                    needs.append(UserNeed(int(access_entity)))
                elif isinstance(access_entity, int):
                    needs.append(UserNeed(access_entity))
            except ValueError:
                needs.append(RoleNeed(access_entity.lower()))
        return needs

    def is_public(self):
        """Check if the record is fully public.

        Explicit permission = `_access` field
        Implicit permission = `restricted` field
        Explicit, when defined, takes precedence over implicit which is
        ignored.
        The record is public when `_access.read` is not defined and
        `restricted` is False.
        """
        has_explicit_perm = current_app.config.get(
            "ILS_RECORDS_EXPLICIT_PERMISSIONS_ENABLED"
        ) and self.record.get("_access", {}).get("read", [])
        if not has_explicit_perm:
            return self.record.get("restricted", False) is False
        return False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loans item permissions."""

import json
from copy import deepcopy
from datetime import timedelta

import arrow
from flask import url_for
from invenio_search import current_search

from tests.helpers import user_login

NEW_LOAN = {
    "document_pid": "CHANGE ME IN EACH TEST",
    "patron_pid": "3",
    "transaction_location_pid": "locid-1",
    "pickup_location_pid": "locid-1",
    "delivery": {"method": "PICKUP"},
    "request_start_date": "2019-09-10",
    "request_expire_date": "2019-09-20",
}

NEW_LOAN_2 = {
    "document_pid": "CHANGE ME IN EACH TEST",
    "patron_pid": "4",
    "transaction_location_pid": "locid-4",  # With Sat and Sun closures
    "pickup_location_pid": "locid-4",
    "delivery": {"method": "PICKUP"},
    "request_start_date": "2024-03-15",  # Friday
    "request_expire_date": "CHANGE-MY-DATE",
}


def test_anonymous_cannot_request_loan(client, json_headers, testdata):
    """Test that anonymous users cannot request a loan."""
    url = url_for("invenio_app_ils_circulation.loan_request")
    res = client.post(url, headers=json_headers, data=json.dumps(NEW_LOAN))
    assert res.status_code == 401


def test_patron_can_request_loan(client, json_headers, users, testdata):
    """Test that a patron can request a loan."""
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-3"
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "PENDING"
    assert loan["document_pid"] == params["document_pid"]
    assert loan["transaction_date"]


def test_patron_can_cancel_loan(client, json_headers, users, testdata, app_config):
    """Test that a patron can cancel its own loan."""
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron3", users)
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-3"
    params["transaction_user_pid"] = str(user.id)

    # Create a Loan
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    loan = res.get_json()

    cancel_url = loan["links"]["actions"]["cancel"]
    meta = loan["metadata"]
    payload = {
        "document_pid": meta["document_pid"],
        "patron_pid": meta["patron_pid"],
        "transaction_location_pid": meta["transaction_location_pid"],
        "transaction_user_pid": meta["transaction_user_pid"],
        "cancel_reason": "USER_CANCEL",
    }
    assert res.status_code == 202

    # Try to cancel a loan that belongs to patron3 as patron1
    user_login(client, "patron1", users)
    cancel_res = client.post(cancel_url, headers=json_headers, data=json.dumps(payload))
    assert cancel_res.status_code == 403

    # Try to cancel a loan that belongs to patron3 as patron3
    user_login(client, "patron3", users)
    cancel_res = client.post(cancel_url, headers=json_headers, data=json.dumps(payload))
    assert cancel_res.status_code == 202


def test_patron_can_request_loan_with_or_without_end_date(
    app, client, json_headers, users, testdata
):
    """Test that a patron can request a loan [with/without] end date."""
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)

    now = arrow.utcnow()
    start_date = (now + timedelta(days=3)).date().isoformat()
    end_date = (now + timedelta(days=15)).date().isoformat()

    # it should succeed when start/end dates provided
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-2"
    params["request_start_date"] = start_date
    params["request_expire_date"] = end_date
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "PENDING"
    assert loan["document_pid"] == params["document_pid"]
    assert loan["request_start_date"] == start_date
    assert loan["request_expire_date"] == end_date
    assert loan["transaction_date"]

    # it should fail when wrong date provided
    params = deepcopy(NEW_LOAN)
    past_end_date = now - timedelta(days=30)
    params["request_expire_date"] = past_end_date.date().isoformat()
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400

    # it should fail when request duration over max
    params = deepcopy(NEW_LOAN)
    days = app.config["ILS_CIRCULATION_LOAN_REQUEST_DURATION_DAYS"]
    past_end_date = now + timedelta(days=days + 1)
    params["request_expire_date"] = past_end_date.date().isoformat()
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400

    # it should fail when request start/end date not provided
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-4"
    params["transaction_user_pid"] = str(user.id)
    del params["request_start_date"]
    del params["request_expire_date"]
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    now = arrow.utcnow()
    start_date = now.date().isoformat()
    days = app.config["ILS_CIRCULATION_LOAN_REQUEST_DURATION_DAYS"]
    end_date = (now + timedelta(days=days)).date().isoformat()
    assert loan["request_start_date"] == start_date
    assert loan["request_expire_date"] == end_date


def test_request_loan_with_or_without_delivery(
    app, client, json_headers, users, testdata
):
    """Test that loan request with or without delivery."""
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)

    previous_dev_methods = app.config["ILS_CIRCULATION_DELIVERY_METHODS"]
    app.config["ILS_CIRCULATION_DELIVERY_METHODS"] = {}
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-12"
    params["transaction_user_pid"] = str(user.id)
    del params["delivery"]
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    app.config["ILS_CIRCULATION_DELIVERY_METHODS"] = {
        "TEST_METHOD1": "",
        "TEST_METHOD2": "",
    }
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-6"
    params["delivery"] = {"method": "TEST_METHOD1"}
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    app.config["ILS_CIRCULATION_DELIVERY_METHODS"] = {"TEST_METHOD": ""}
    params = deepcopy(NEW_LOAN)
    params["transaction_user_pid"] = str(user.id)
    del params["delivery"]
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    app.config["ILS_CIRCULATION_DELIVERY_METHODS"] = {"TEST_METHOD": ""}
    params = deepcopy(NEW_LOAN)
    params["delivery"] = {"method": "NON_EXISTING_METHOD"}
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    # restore
    app.config["ILS_CIRCULATION_DELIVERY_METHODS"] = previous_dev_methods


def test_request_loan_with_active_loan_or_loan_request(
    app, client, json_headers, users, testdata
):
    """Test that a patron cannot request a loan with already an active loan or
    a loan request on the same document.
    """
    # Try to request document with already a loan request
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-3"
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    current_search.flush_and_refresh(index="*")
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400

    # Try to request document with already an active loan
    user_login(client, "librarian", users)
    url = url_for("invenio_app_ils_circulation.loan_checkout")
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-3"
    params["transaction_user_pid"] = str(users["librarian"].id)
    params["item_pid"] = dict(type="pitmid", value="itemid-60")
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    current_search.flush_and_refresh(index="*")
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)
    params = deepcopy(NEW_LOAN)
    params["document_pid"] = "docid-3"
    params["transaction_user_pid"] = str(user.id)
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400


def test_request_loan_minimum_days(app, client, json_headers, users, testdata):
    """Test that a patron can only request a loan after a set minimum days"""
    url = url_for("invenio_app_ils_circulation.loan_request")
    user = user_login(client, "patron1", users)
    app.config["ILS_CIRCULATION_LOAN_REQUEST_OFFSET"] = 4
    params = deepcopy(NEW_LOAN)

    now = arrow.utcnow()
    start_date = now.date().isoformat()
    params["request_start_date"] = start_date
    params["transaction_user_pid"] = str(user.id)

    params["document_pid"] = "docid-4"
    end_date = (now + timedelta(days=2)).date().isoformat()  # FAIL
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    params["document_pid"] = "docid-5"
    end_date = (now - timedelta(days=2)).date().isoformat()  # FAIL
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    params["document_pid"] = "docid-6"
    end_date = now.date().isoformat()  # FAIL
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    params["document_pid"] = "docid-7"
    end_date = (now + timedelta(days=4)).date().isoformat()  # PASS
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "PENDING"
    assert loan["document_pid"] == params["document_pid"]
    assert loan["transaction_date"]

    params["document_pid"] = "docid-8"
    end_date = (now + timedelta(days=8)).date().isoformat()  # PASS
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "PENDING"
    assert loan["document_pid"] == params["document_pid"]
    assert loan["transaction_date"]

    app.config["ILS_CIRCULATION_LOAN_REQUEST_OFFSET"] = -3

    params["document_pid"] = "docid-9"
    end_date = now.date().isoformat()  # FAIL
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    params["document_pid"] = "docid-10"
    end_date = (now - timedelta(days=2)).date().isoformat()  # FAIL
    params["request_expire_date"] = end_date
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    app.config["ILS_CIRCULATION_LOAN_REQUEST_OFFSET"] = 2
    # By-pass default location validation for testing closures on a different location
    location_validator, app.config["CIRCULATION_TRANSACTION_LOCATION_VALIDATOR"] = (
        app.config["CIRCULATION_TRANSACTION_LOCATION_VALIDATOR"],
        lambda _: True,
    )
    # Request Date is Friday
    params = deepcopy(NEW_LOAN_2)
    params["transaction_user_pid"] = str(user.id)

    params["document_pid"] = "docid-11"
    # Monday
    params["request_expire_date"] = "2024-03-18"  # FAIL
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 400
    assert res.get_json()["message"] == "Validation error."

    params["document_pid"] = "docid-12"
    # Tuesday
    params["request_expire_date"] = "2024-03-19"  # PASS
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202
    loan = res.get_json()["metadata"]
    assert loan["state"] == "PENDING"
    assert loan["document_pid"] == params["document_pid"]
    assert loan["transaction_date"]

    # Set the config back to its original state
    app.config["CIRCULATION_TRANSACTION_LOCATION_VALIDATOR"] = location_validator



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS Document Request Document loader JSON schema."""

from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from marshmallow import EXCLUDE, fields


class DocumentRequestDocumentSchemaV1(RecordMetadataSchemaJSONV1):
    """Document Request Document schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    document_pid = fields.Str(required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan item resolver."""

from invenio_circulation.api import Loan


def test_loan_document_resolver(app, testdata):
    """Test item resolving from loan."""
    loan_pid = testdata["loans"][1]["pid"]
    loan = Loan.get_record_by_pid(loan_pid)
    loan = loan.replace_refs()
    assert loan["document"]["pid"] == loan["document_pid"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio app ils signals."""

from blinker import Namespace

_signals = Namespace()

record_viewed = _signals.signal("record-viewed")
file_downloaded = _signals.signal("file-downloaded")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Circulation notifications."""
from celery import shared_task
from celery.utils.log import get_task_logger
from flask import current_app

from invenio_app_ils.circulation.notifications.api import (
    send_expiring_loan_reminder_notification,
    send_loan_notification,
    send_loan_overdue_reminder_notification,
)
from invenio_app_ils.circulation.search import (
    get_all_expiring_loans,
    get_all_overdue_loans,
)
from invenio_app_ils.circulation.utils import circulation_overdue_loan_days

celery_logger = get_task_logger(__name__)


@shared_task
def send_overdue_loans_notification_reminder():
    """Send notification message for loans that are overdue every X days."""
    days = current_app.config["ILS_CIRCULATION_NOTIFICATION_OVERDUE_REMINDER_INTERVAL"]
    for hit in get_all_overdue_loans().scan():
        loan = hit.to_dict()
        days_ago = circulation_overdue_loan_days(loan)
        if days_ago % days == 0:
            send_loan_overdue_reminder_notification(loan, days_ago)


@shared_task
def send_expiring_loans_notification_reminder():
    """Send notification for loans that will expire in X days."""
    days = current_app.config["ILS_CIRCULATION_LOAN_WILL_EXPIRE_DAYS"]
    for hit in get_all_expiring_loans(days).scan():
        loan = hit.to_dict()
        send_expiring_loan_reminder_notification(loan, days)


@shared_task
def send_loan_end_date_updated_notification(loan):
    """Sends an notification to the patron whose loan was updated."""
    send_loan_notification(
        action="extend",
        loan=loan,
        is_manually_triggered=False,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""EItem indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_indexer.api import RecordIndexer

from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE
from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.proxies import current_app_ils

from .api import EITEM_PID_TYPE


@shared_task(ignore_result=True)
def index_referenced_records(eitem):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()
    indexed = dict(pid_type=EITEM_PID_TYPE, record=eitem)

    # fetch and index the document
    document_cls = current_app_ils.document_record_cls
    document = document_cls.get_record_by_pid(eitem["document_pid"])
    referenced = [dict(pid_type=DOCUMENT_PID_TYPE, record=document)]

    indexer.index(indexed, referenced)


class EItemIndexer(RecordIndexer):
    """Indexer class for EItem record."""

    def index(self, eitem, arguments=None, **kwargs):
        """Index an EItem."""
        super().index(eitem)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((eitem,), eta=eta)


@shared_task(ignore_result=True)
def index_eitem_after_files_changed(bucket_id):
    """Index EItem on file changed."""
    search_cls = current_app_ils.eitem_search_cls
    results = search_cls().search_by_bucket_id(bucket_id)
    if len(results) == 1:
        hit = results[0]
        record = current_app_ils.eitem_record_cls.get_record_by_pid(hit.pid)
        current_app_ils.eitem_indexer.index(record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""APIs to retrieve records relations."""

from invenio_app_ils.records.jsonresolvers.api import pick
from invenio_app_ils.relations.api import (
    PARENT_CHILD_RELATION_TYPES,
    SEQUENCE_RELATION_TYPES,
    SIBLINGS_RELATION_TYPES,
    ParentChildRelation,
    SequenceRelation,
    SiblingsRelation,
)

from .api import RecordRelationsExtraMetadata as RelationsExtraMetadata


class RelationObjectBuilderMixin(object):
    """Relations object builder."""

    def build_relations_object(self, pid_value, pid_type, relation_name, **kwargs):
        """Build and return each object of `relations` field."""
        r = {
            "pid_value": pid_value,
            "pid_type": pid_type,
            "relation_type": relation_name,
        }
        r.update(kwargs)
        return r

    def get_relevant_fields_from(self, record):
        """Get relevant fields from the related record."""
        fields = pick(
            record,
            "title",
            "edition",
            "publication_year",
            "languages",
            "document_type",
            "mode_of_issuance",
        )
        return dict(record_metadata=fields)


class ParentChildRetriever(RelationObjectBuilderMixin):
    """Retrieve relations of type ParentChild."""

    def __init__(self, child_record):
        """Constructor."""
        self.child_record = child_record

    def _build_relation_obj(self, child, parent_pid, relation_type):
        """Return the relation object with any extra metadata."""
        pid_value = parent_pid.pid_value
        pid_type = parent_pid.pid_type

        r = self.build_relations_object(pid_value, pid_type, relation_type)

        metadata = RelationsExtraMetadata.get_extra_metadata_from(
            child, relation_type, parent_pid.pid_value, parent_pid.pid_type
        )
        # copy the extra metadata into the relation, if any
        r.update(metadata)

        # fetch the parent to retrieve useful metadata (title, ...)
        from invenio_app_ils.records.api import IlsRecord

        parent = IlsRecord.get_record_by_pid(pid_value, pid_type=pid_type)

        # copy relevant fields from parent into the relation
        relevant_fields = self.get_relevant_fields_from(parent)
        r.update(relevant_fields)
        return r

    def get(self):
        """Return only relations that this record has with parents.

        Given that in ILS system parent records (e.g. journal, periodicals)
        can have a very high number of children but only a few parents,
        only relations with the parents are returned.
        """
        relations = {}
        for relation_type in PARENT_CHILD_RELATION_TYPES:
            pcr = ParentChildRelation(relation_type)
            name = relation_type.name

            for parent_pid in pcr.get_parents_of(self.child_record.pid):
                r = self._build_relation_obj(self.child_record, parent_pid, name)
                relations.setdefault(name, [])
                relations[name].append(r)
        return relations


class SiblingsRetriever(RelationObjectBuilderMixin):
    """Retrieve relations of type Siblings."""

    def __init__(self, record):
        """Constructor."""
        self.record = record

    @staticmethod
    def _get_extra_metadata(record, sibling, relation_type):
        """Retrieve any extra metadata in the current record or sibling."""
        record_pid = record.pid
        sibling_pid = sibling.pid
        # we don't know it extra metadata are stored in this record or in the
        # sibling. Need to check both.
        metadata = {}

        # get any extra metadata stored in this record for the relation with
        # the sibling
        rec_metadata = RelationsExtraMetadata.get_extra_metadata_from(
            record, relation_type, sibling_pid.pid_value, sibling_pid.pid_type
        )
        metadata.update(rec_metadata)

        # get any extra metadata stored in the sibling for the relation with
        # this record
        sibling_metadata = RelationsExtraMetadata.get_extra_metadata_from(
            sibling, relation_type, record_pid.pid_value, record_pid.pid_type
        )
        if (
            sibling_metadata
            and "pid_value" in sibling_metadata
            and "pid_type" in sibling_metadata
        ):
            # the extra metadata in the sibling has pid_value and pid_type of
            # this record. Replace it with the sibling ones.
            sibling_metadata["pid_value"] = sibling_pid.pid_value
            sibling_metadata["pid_type"] = sibling_pid.pid_type

        metadata.update(sibling_metadata)
        return metadata

    def _build_relation_obj(self, sibling_pid, relation_type):
        """Return the relation object with metadata."""
        pid_value = sibling_pid.pid_value
        pid_type = sibling_pid.pid_type

        r = self.build_relations_object(pid_value, pid_type, relation_type)

        # fetch the sibling to retrieve useful metadata (title, ...) and also
        # any optional extra metadata for this relation
        from invenio_app_ils.records.api import IlsRecord

        sibling = IlsRecord.get_record_by_pid(pid_value, pid_type=pid_type)

        metadata = self._get_extra_metadata(self.record, sibling, relation_type)
        # copy the extra metadata into the relation, if any
        r.update(metadata)

        # copy relevant fields from sibling into the relation
        relevant_fields = self.get_relevant_fields_from(sibling)
        r.update(relevant_fields)
        return r

    def get(self):
        """Get all sibling relations with the current record."""
        relations = {}
        for relation_type in SIBLINGS_RELATION_TYPES:
            sr = SiblingsRelation(relation_type)
            name = relation_type.name

            for sibling_pid in sr.all(self.record.pid):
                r = self._build_relation_obj(sibling_pid, name)
                relations.setdefault(name, [])
                relations[name].append(r)

            # Pre-requisite: fields being sorted should be of the same type.
            # Sort by the keys available in the data in order
            if relation_type.sort_by:
                relations.get(name, []).sort(
                    key=lambda rec: [
                        rec.get("record_metadata", {}).get(key)
                        for key in relation_type.sort_by
                    ]
                )

        return relations


class SequenceRetriever(RelationObjectBuilderMixin):
    """Retrieve relations of type Sequence."""

    ORDER_FIELD_NAME = "relation_order"
    ORDER_VALUE_NEXT = "continues"
    ORDER_VALUE_PREVIOUS = "is_continued_by"

    def __init__(self, record):
        """Constructor."""
        self.record = record

    def _build_relation_obj(self, related_pid, relation_type):
        """Return the relation object with metadata."""
        pid_value = related_pid.pid_value
        pid_type = related_pid.pid_type

        r = self.build_relations_object(pid_value, pid_type, relation_type)

        # fetch the sequence to retrieve useful metadata (title, ...)
        from invenio_app_ils.records.api import IlsRecord

        seq_rec = IlsRecord.get_record_by_pid(pid_value, pid_type=pid_type)

        # copy relevant fields from the sequence record into the relation
        relevant_fields = self.get_relevant_fields_from(seq_rec)
        r.update(relevant_fields)
        return r

    def get(self):
        """Get all sequence relations with the current record."""
        relations = {}
        for relation_type in SEQUENCE_RELATION_TYPES:
            sqr = SequenceRelation(relation_type)
            name = relation_type.name

            # pid_5 --- is previous of ---> [next_6, next_7, ...]
            for pid in sqr.pid_is_previous(self.record.pid):
                r = self._build_relation_obj(pid, name)
                r[self.ORDER_FIELD_NAME] = self.ORDER_VALUE_PREVIOUS
                relations.setdefault(name, [])
                relations[name].append(r)

            # pid_5 --- is next of ---> [previous_3, previous_4, ...]
            for pid in sqr.pid_is_next(self.record.pid):
                r = self._build_relation_obj(pid, name)
                r[self.ORDER_FIELD_NAME] = self.ORDER_VALUE_NEXT
                relations.setdefault(name, [])
                relations[name].append(r)
        return relations


def get_relations(record):
    """Get all relations for the given record."""
    relations = {}

    pc_rels = ParentChildRetriever(record).get()
    relations.update(pc_rels)

    sibl_rels = SiblingsRetriever(record).get()
    relations.update(sibl_rels)

    seq_rels = SequenceRetriever(record).get()
    relations.update(seq_rels)

    return relations



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest fixtures and plugins for the API application."""

import tempfile

import pytest
from invenio_app.factory import create_api
from invenio_circulation.api import Loan
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search

from invenio_app_ils.acquisition.api import ORDER_PID_TYPE, Order
from invenio_app_ils.document_requests.api import (
    DOCUMENT_REQUEST_PID_TYPE,
    DocumentRequest,
)
from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE, Document
from invenio_app_ils.eitems.api import EITEM_PID_TYPE, EItem
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE, BorrowingRequest
from invenio_app_ils.internal_locations.api import (
    INTERNAL_LOCATION_PID_TYPE,
    InternalLocation,
)
from invenio_app_ils.items.api import ITEM_PID_TYPE, Item
from invenio_app_ils.locations.api import LOCATION_PID_TYPE, Location
from invenio_app_ils.providers.api import PROVIDER_PID_TYPE, Provider
from invenio_app_ils.series.api import SERIES_PID_TYPE, Series
from tests.helpers import (
    document_ref_builder,
    internal_location_ref_builder,
    load_json_from_datadir,
    mint_record_pid,
)


@pytest.fixture(scope="module")
def app_config(app_config):
    """Get app config."""
    tests_config = {
        "REST_CSRF_ENABLED": False,
        "ACCOUNTS_SESSION_REDIS_URL": "",  # in-memory
        "RATELIMIT_GUEST_USER": "1000 per minute",
        "RATELIMIT_AUTHENTICATED_USER": "1000 per minute",
        "CIRCULATION_TRANSACTION_USER_VALIDATOR": lambda x: True,
        "EXTEND_LOANS_LOCATION_UPDATED": False,
        "ILS_NOTIFICATIONS_TEMPLATES": {"footer": "footer.html"},
    }
    app_config.update(tests_config)
    return app_config


@pytest.fixture(scope="module")
def create_app():
    """Create test app."""
    return create_api


@pytest.fixture()
def json_headers():
    """JSON headers."""
    return [
        ("Content-Type", "application/json"),
        ("Accept", "application/json"),
    ]


def _create_records(db, objs, cls, pid_type):
    """Create records and index."""
    recs = []
    for obj in objs:
        record = cls.create(obj)
        mint_record_pid(pid_type, "pid", record)
        recs.append(record)
    db.session.commit()
    return recs


@pytest.fixture()
def testdata(app, db, search_clear, users):
    """Create, index and return test data."""
    data = load_json_from_datadir("locations.json")
    locations = _create_records(db, data, Location, LOCATION_PID_TYPE)

    data = load_json_from_datadir("internal_locations.json")
    int_locs = _create_records(db, data, InternalLocation, INTERNAL_LOCATION_PID_TYPE)

    data = load_json_from_datadir("series.json")
    series = _create_records(db, data, Series, SERIES_PID_TYPE)

    data = load_json_from_datadir("documents.json")
    documents = _create_records(db, data, Document, DOCUMENT_PID_TYPE)

    data = load_json_from_datadir("items.json")
    items = _create_records(db, data, Item, ITEM_PID_TYPE)

    data = load_json_from_datadir("eitems.json")
    eitems = _create_records(db, data, EItem, EITEM_PID_TYPE)

    data = load_json_from_datadir("loans.json")
    loans = _create_records(db, data, Loan, CIRCULATION_LOAN_PID_TYPE)

    data = load_json_from_datadir("acq_providers.json")
    acq_providers = _create_records(db, data, Provider, PROVIDER_PID_TYPE)

    data = load_json_from_datadir("acq_orders.json")
    acq_orders = _create_records(db, data, Order, ORDER_PID_TYPE)

    data = load_json_from_datadir("ill_providers.json")
    ill_providers = _create_records(db, data, Provider, PROVIDER_PID_TYPE)

    data = load_json_from_datadir("ill_borrowing_requests.json")
    ill_brw_reqs = _create_records(
        db, data, BorrowingRequest, BORROWING_REQUEST_PID_TYPE
    )

    data = load_json_from_datadir("document_requests.json")
    doc_reqs = _create_records(db, data, DocumentRequest, DOCUMENT_REQUEST_PID_TYPE)

    # index
    ri = RecordIndexer()
    for rec in (
        locations
        + int_locs
        + series
        + documents
        + items
        + eitems
        + loans
        + doc_reqs
        + acq_providers
        + acq_orders
        + ill_providers
        + ill_brw_reqs
    ):
        ri.index(rec)

    current_search.flush_and_refresh(index="*")

    return {
        "document_requests": doc_reqs,
        "documents": documents,
        "internal_locations": int_locs,
        "items": items,
        "eitems": eitems,
        "loans": loans,
        "locations": locations,
        "series": series,
        "acq_providers": acq_providers,
        "acq_orders": acq_orders,
        "ill_providers": ill_providers,
        "ill_brw_reqs": ill_brw_reqs,
    }


@pytest.fixture()
def testdata_most_loaned(db, testdata):
    """Create, index and return test data for most loans tests."""
    most_loaned = load_json_from_datadir("loans_most_loaned.json")
    recs = _create_records(db, most_loaned, Loan, CIRCULATION_LOAN_PID_TYPE)

    ri = RecordIndexer()
    for rec in recs:
        ri.index(rec)

    current_search.flush_and_refresh(index="loans")

    return {
        "locations": testdata["locations"],
        "internal_locations": testdata["internal_locations"],
        "documents": testdata["documents"],
        "items": testdata["items"],
        "loans": most_loaned,
        "series": testdata["series"],
    }


@pytest.fixture()
def item_record(app):
    """Fixture to return an Item payload."""
    return {
        "pid": "itemid-1",
        "document_pid": "docid-1",
        "document": {"$ref": document_ref_builder(app, "itemid-1")},
        "barcode": "123456789-1",
        "title": "Test item x",
        "internal_location_pid": "ilocid-1",
        "internal_location": {"$ref": internal_location_ref_builder(app, "itemid-1")},
        "medium": "NOT_SPECIFIED",
        "status": "CAN_CIRCULATE",
        "circulation_restriction": "NO_RESTRICTION",
    }


@pytest.fixture()
def loan_params():
    """Params for API REST payload."""
    return dict(
        transaction_user_pid="4",
        patron_pid="1",
        document_pid="docid-1",
        item_pid=dict(type="pitmid", value="itemid-2"),
        transaction_location_pid="locid-1",
        transaction_date="2018-02-01T09:30:00+02:00",
    )


@pytest.fixture()
def bucket(bucket_from_dir):
    """Create temporary bucket fixture."""
    with tempfile.TemporaryDirectory(prefix="ils-test-") as temp_dir:
        bucket = bucket_from_dir(temp_dir)
        yield bucket


@pytest.fixture()
def with_access(app):
    """Enable explicit permission check (`_access`)."""
    app.config["ILS_RECORDS_EXPLICIT_PERMISSIONS_ENABLED"] = True
    yield
    app.config["ILS_RECORDS_EXPLICIT_PERMISSIONS_ENABLED"] = False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loan update."""
import json
from copy import deepcopy
from datetime import date, timedelta

from flask import url_for
from invenio_circulation.api import Loan

from tests.helpers import user_login, user_logout


def _url_loan(pid_value):
    return url_for("invenio_records_rest.loanid_item", pid_value=pid_value)


def _post_loan_update(
    client,
    json_headers,
    pid_value,
    start_date=None,
    end_date=None,
    request_start_date=None,
    request_expire_date=None,
):
    url = url_for(
        "invenio_app_ils_circulation.loanid_update_dates", pid_value=pid_value
    )
    data = {}
    if start_date:
        data["start_date"] = start_date
    if end_date:
        data["end_date"] = end_date
    if request_start_date:
        data["request_start_date"] = request_start_date
    if request_expire_date:
        data["request_expire_date"] = request_expire_date
    res = client.post(url, headers=json_headers, data=json.dumps(data))
    return res


def _load_result(res):
    return json.loads(res.data.decode("utf-8"))


def _today(dt=0):
    return (date.today() + timedelta(days=dt)).strftime("%Y-%m-%d")


def test_loan_access_permission(client, json_headers, users, testdata):
    """
    Test that a patron should not be able to update their loan;
    and a fortiori, other people's.
    """
    loan = testdata["loans"][0]
    user = user_login(client, "patron1", users)
    url = _url_loan(loan["pid"])
    res = client.get(url, headers=json_headers)
    assert res.status_code == 200  # Can access their own loan
    metadata = _load_result(res)["metadata"]
    metadata["transaction_user_pid"] = str(user.id)
    res = client.put(url, headers=json_headers, data=json.dumps(metadata))
    assert res.status_code == 403  # Cannot modify the loan


def test_loan_update_date(client, json_headers, users, testdata):
    """Test the edition of the dates on a loan."""
    pid = testdata["loans"][4]["pid"]  # Item on loan

    res = _post_loan_update(client, json_headers, pid)
    assert res.status_code == 401

    # Patron cannot update loan
    user_login(client, "patron1", users)
    res = _post_loan_update(client, json_headers, pid)
    assert res.status_code == 403
    user_logout(client)

    # Librarian can
    user_login(client, "librarian", users)
    old_loan = _load_result(client.get(_url_loan(pid), headers=json_headers))
    res = _post_loan_update(client, json_headers, pid)
    assert res.status_code == 202
    new_loan = _load_result(res)
    assert old_loan["updated"] < new_loan["updated"]  # Liveness check

    # Update both values
    start_date = "2020-01-01"
    end_date = _today(+1)
    res = _post_loan_update(
        client, json_headers, pid, start_date=start_date, end_date=end_date
    )
    assert res.status_code == 202
    new_loan_meta = _load_result(res)["metadata"]
    assert new_loan_meta["start_date"] == start_date
    assert new_loan_meta["end_date"] == end_date

    # Start date after today
    start_date = _today(+2)
    end_date = _today(+3)
    res = _post_loan_update(
        client, json_headers, pid, start_date=start_date, end_date=end_date
    )
    assert res.status_code == 400

    # No relative constraints on non-active loans
    pid = testdata["loans"][3]["pid"]  # Pending
    res = _post_loan_update(
        client,
        json_headers,
        pid,
        request_start_date=start_date,
        request_expire_date=end_date,
    )
    assert res.status_code == 202
    new_loan_meta = _load_result(res)["metadata"]
    assert new_loan_meta["request_start_date"] == start_date
    assert new_loan_meta["request_expire_date"] == end_date

    # Negative date range
    start_date = "2000-02-01"
    end_date = "2000-01-01"
    res = _post_loan_update(
        client,
        json_headers,
        pid,
        request_start_date=start_date,
        request_expire_date=end_date,
    )
    assert res.status_code == 400

    # Illegal combination of parameters
    start_date = _today(-2)
    end_date = _today(+2)
    res = _post_loan_update(
        client,
        json_headers,
        pid,
        start_date=start_date,
        end_date=end_date,
        request_start_date=start_date,
        request_expire_date=end_date,
    )
    assert res.status_code == 400



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Location indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_indexer.api import RecordIndexer

from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.internal_locations.api import INTERNAL_LOCATION_PID_TYPE
from invenio_app_ils.items.api import ITEM_PID_TYPE
from invenio_app_ils.locations.api import LOCATION_PID_TYPE
from invenio_app_ils.proxies import current_app_ils


def get_internal_locations(location_pid):
    """Get referenced internal location."""
    referenced = []
    intloc_search_cls = current_app_ils.internal_location_search_cls
    intloc_record_cls = current_app_ils.internal_location_record_cls
    for iloc in (
        intloc_search_cls().search_by_location_pid(location_pid=location_pid).scan()
    ):
        intloc = intloc_record_cls.get_record_by_pid(iloc["pid"])
        referenced.append(dict(pid_type=INTERNAL_LOCATION_PID_TYPE, record=intloc))
    return referenced


def get_items(location_pid):
    """Get referenced items."""
    referenced = []
    item_search_cls = current_app_ils.item_search_cls
    item_record_cls = current_app_ils.item_record_cls
    for item in (
        item_search_cls().search_by_location_pid(location_pid=location_pid).scan()
    ):
        item = item_record_cls.get_record_by_pid(item["pid"])
        referenced.append(dict(pid_type=ITEM_PID_TYPE, record=item))
    return referenced


@shared_task(ignore_result=True)
def index_referenced_records(location):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()

    location_pid = location["pid"]
    indexed = dict(pid_type=LOCATION_PID_TYPE, record=location)

    indexer.index(indexed, get_internal_locations(location_pid))
    indexer.index(indexed, get_items(location_pid))


class LocationIndexer(RecordIndexer):
    """Indexer class for Location record."""

    def index(self, location, arguments=None, **kwargs):
        """Index an Location."""
        super().index(location)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((location,), eta=eta)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the InternalLocation referenced in the Item."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.internal_locations.api import InternalLocation
from invenio_app_ils.items.api import Item
from invenio_app_ils.records.jsonresolvers.api import (
    get_field_value_for_record as get_field_value,
)
from invenio_app_ils.records.jsonresolvers.api import get_pid_or_default, pick

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Internal Location for an Item record."""
    from flask import current_app

    @get_pid_or_default(default_value=dict())
    def get_internal_location(internal_location_pid):
        """Return the InternalLocation record."""
        internal_location = InternalLocation.get_record_by_pid(internal_location_pid)
        del internal_location["$schema"]
        if "notes" in internal_location:
            del internal_location["notes"]

        return internal_location

    def internal_location_resolver(item_pid):
        """Return the IntLoc record for the given Item or raise."""
        internal_loc_pid = get_field_value(Item, item_pid, "internal_location_pid")
        return get_internal_location(internal_loc_pid)

    url_map.add(
        Rule(
            "/api/resolver/items/<item_pid>/internal-location",
            endpoint=internal_location_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loan extend."""

import json
from copy import deepcopy
from datetime import timedelta

import arrow
from flask import url_for
from invenio_circulation.api import Loan
from invenio_db import db

from invenio_app_ils.circulation.api import circulation_default_loan_duration_for_item
from invenio_app_ils.items.api import Item
from tests.helpers import user_login


def _checkout_loan_pid1(client, json_headers, users, loan_params):
    """Create an ongoing loan."""

    def _checkout(loan_pid, params):
        """Perform checkout action."""
        user_login(client, "librarian", users)
        checkout_url = url_for(
            "invenio_circulation_loan_actions.loanid_actions",
            pid_value=loan_pid,
            action="checkout",
        )
        resp = client.post(checkout_url, headers=json_headers, data=json.dumps(params))
        assert resp.status_code == 202
        return resp.get_json()

    loan_pid = "loanid-1"
    params = deepcopy(loan_params)
    params["document_pid"] = "docid-1"
    params["item_pid"]["value"] = "itemid-2"

    return _checkout(loan_pid, params)


def test_loan_extend_permissions(client, json_headers, users, testdata, loan_params):
    """Test loan can be extended."""
    params = deepcopy(loan_params)
    del params["transaction_date"]
    loan = _checkout_loan_pid1(client, json_headers, users, params)

    tests = [
        ("admin", 202),
        ("librarian", 202),
        ("patron1", 202),
        ("patron3", 403),
    ]

    for username, expected_resp_code in tests:
        extend_url = loan["links"]["actions"]["extend"]

        user_login(client, username, users)
        res = client.post(
            extend_url,
            headers=json_headers,
            data=json.dumps(params),
        )
        assert res.status_code == expected_resp_code


def test_loan_extension_end_date(
    app, client, json_headers, users, testdata, loan_params
):
    """Test loan end date after extension."""
    params = deepcopy(loan_params)
    del params["transaction_date"]
    record = _checkout_loan_pid1(client, json_headers, users, params)
    extend_url = record["links"]["actions"]["extend"]
    loan = record["metadata"]

    item = Item.get_record_by_pid(loan["item_pid"]["value"])
    item_loan_duration = circulation_default_loan_duration_for_item(item)

    user_login(client, "patron1", users)

    def _set_loan_end_date(loan_pid, new_end_date):
        """Set loan end date."""
        loan = Loan.get_record_by_pid(loan_pid)
        loan["end_date"] = new_end_date.date().isoformat()
        loan.commit()
        db.session.commit()

    def test_extension_end_date_loan_not_overdue(loan_pid):
        """Test that new end date is added to the current end date."""
        now = arrow.utcnow()
        new_end_date = now + timedelta(days=10)  # loan is not overdue
        _set_loan_end_date(loan_pid, new_end_date)

        expected_end_date = new_end_date + item_loan_duration

        res = client.post(extend_url, headers=json_headers, data=json.dumps(params))
        assert res.status_code == 202
        new_loan = res.get_json()["metadata"]
        assert new_loan["end_date"] == expected_end_date.date().isoformat()

    def test_extension_end_date_loan_overdue(loan_pid):
        """Test that new end date is added to today."""
        now = arrow.utcnow()

        new_end_date = now - timedelta(days=2)  # loan is overdue since 2 days
        _set_loan_end_date(loan_pid, new_end_date)

        expected_end_date = now + item_loan_duration
        res = client.post(extend_url, headers=json_headers, data=json.dumps(params))
        assert res.status_code == 202
        new_loan = res.get_json()["metadata"]
        # loan new end date should start from now
        assert new_loan["end_date"] == expected_end_date.date().isoformat()

    test_extension_end_date_loan_not_overdue(loan["pid"])
    test_extension_end_date_loan_overdue(loan["pid"])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# Invenio-App-Ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Facets and factories for result filtering and aggregation."""

import arrow
from flask import current_app
from invenio_search.engine import dsl


def keyed_range_filter(field, range_query, **kwargs):
    """Create a range filter.

    :param field: Field name.
    :param range_query: Dictionary with available keys and their range options.
    """

    def inner(values):
        args = {}
        for range_key, mappings in range_query.items():
            if range_key in values:
                for key, value in mappings.items():
                    args[key] = value

        args.update(kwargs.copy())
        return dsl.query.Range(**{field: args})

    return inner


def default_value_when_missing_filter(field, missing_val):
    """Create a custom exists filter.

    :param field: Field name.
    :missing_val
    :returns: Function that returns the Terms query.
    """

    def inner(values):
        if missing_val in values:
            return dsl.query.Bool(**{"must_not": {"exists": {"field": field}}})
        else:
            return dsl.Q("terms", **{field: values})

    return inner


def exists_value_filter(field, filter_value):
    """Create a custom filter that filters by existing value.

    :param field: Field name.
    :param filter_value: Filter value.
    """

    def inner(values):
        if filter_value in values:
            return dsl.query.Bool(**{"must": {"exists": {"field": field}}})
        else:
            return dsl.query.Bool(**{"must_not": {"exists": {"field": field}}})

    return inner


def overdue_loans_filter(field):
    """Create a custom filter for ongoing loans.

    :param field: Field to filter.
    :param range_query: Dictionary with available keys and their range options.
    """

    def inner(values):
        range_query = {
            "Overdue": {"lt": str(arrow.utcnow().date())},
            "Upcoming return": {
                "lte": str(
                    current_app.config["CIRCULATION_POLICIES"][
                        "upcoming_return_range"
                    ]().date()
                ),
                "gte": str(arrow.utcnow().date()),
            },
        }

        args = {}
        for range_key, mappings in range_query.items():
            if range_key in values:
                for key, value in mappings.items():
                    args[key] = value
        range_query = dsl.query.Range(**{field: args})
        terms_query = dsl.Q(
            "terms", **{"state": current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]}
        )

        return dsl.query.Bool(must=[range_query, terms_query])

    return inner


def overdue_agg():
    """Create a custom aggregation with dynamic dates."""
    return dict(
        filter=dict(
            terms=dict(state=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"])
        ),
        aggs=dict(
            end_date=dict(
                range=dict(
                    field="end_date",
                    ranges=[
                        {"key": "Overdue", "to": str((arrow.utcnow()).date())},
                        {
                            "key": "Upcoming return",
                            "from": str(arrow.utcnow().date()),
                            "to": str(
                                current_app.config["CIRCULATION_POLICIES"][
                                    "upcoming_return_range"
                                ]().date()
                            ),
                        },
                    ],
                )
            )
        ),
    )


def date_range_filter(field, comparator):
    """Create a range filter.

    :param field: Field name.
    :param comparator: Comparison we want with the supplied date.
    """

    def inner(values):
        try:
            input_date = str(arrow.get(values[0]).date())
        except arrow.parser.ParserError:
            raise ValueError("Input should be a date")
        return dsl.RangeField(**{field: {comparator: input_date}})

    return inner



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Loan custom serializer functions."""

from flask import current_app
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError

from invenio_app_ils.circulation.utils import circulation_overdue_loan_days
from invenio_app_ils.errors import PatronNotFoundError
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.jsonresolvers.api import pick


def field_is_overdue(metadata):
    """Calculate if the loan is overdue."""
    metadata["is_overdue"] = False
    is_loan_active = (
        metadata["state"] in current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
    )
    if is_loan_active and "end_date" in metadata:
        metadata["is_overdue"] = circulation_overdue_loan_days(metadata) > 0


def field_pickup_location(metadata):
    """Get the pickup location object and add it a property."""
    pickup_location_pid = metadata.get("pickup_location_pid")
    if not pickup_location_pid:
        return
    Location = current_app_ils.location_record_cls
    try:
        pickup_location = Location.get_record_by_pid(pickup_location_pid)
    except PIDDeletedError:
        metadata["pickup_location"] = {"name": "This location was deleted."}
        return
    except PIDDoesNotExistError:
        metadata["pickup_location"] = {"name": "Location PID invalid."}
        return
    metadata["pickup_location"] = pick(pickup_location, "name")


def field_transaction_location(metadata):
    """Get the transaction location object and add it a property."""
    transaction_location_pid = metadata.get("transaction_location_pid")
    if not transaction_location_pid:
        return
    Location = current_app_ils.location_record_cls
    try:
        transaction_location = Location.get_record_by_pid(transaction_location_pid)
    except PIDDeletedError:
        metadata["transaction_location"] = {"name": "This location was deleted."}
        return
    except PIDDoesNotExistError:
        metadata["transaction_location"] = {"name": "Location PID invalid."}
        return
    metadata["transaction_location"] = pick(transaction_location, "name")


def field_transaction_user(metadata):
    """Get the transaction location object and add it a property."""
    transaction_user_pid = metadata.get("transaction_user_pid")
    if not transaction_user_pid:
        return
    Patron = current_app_ils.patron_cls
    try:
        transaction_user = Patron.get_patron(transaction_user_pid)
    except (PIDDeletedError, PatronNotFoundError):
        return
    metadata["transaction_user"] = pick(transaction_user, "name")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Literature custom serializer functions."""

from flask import current_app


def field_cover_metadata(metadata):
    """Build urls for literature covers."""
    url_builder = current_app.config.get("ILS_LITERATURE_COVER_URLS_BUILDER")
    cover_metadata = metadata.get("cover_metadata", {})
    cover_metadata["urls"] = url_builder(metadata)
    metadata["cover_metadata"] = cover_metadata



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests items CRUD."""

from copy import deepcopy

import pytest

from invenio_app_ils.circulation.search import get_active_loan_by_item_pid
from invenio_app_ils.errors import (
    DocumentNotFoundError,
    InternalLocationNotFoundError,
    ItemHasPastLoansError,
)
from invenio_app_ils.items.api import ITEM_PID_TYPE, Item


def test_item_create(app, testdata):
    """Test creation of an item."""
    item = Item.create(
        dict(
            pid="itemid-99",
            document_pid="docid-1",
            internal_location_pid="ilocid-4",
            created_by=dict(type="script", value="demo"),
            barcode="cm-348048",
            status="CAN_CIRCULATE",
            circulation_restriction="NO_RESTRICTION",
            medium="PAPER",
        )
    )
    assert "$schema" in item
    assert "document" in item and "$ref" in item["document"]
    assert item["barcode"] == "CM-348048"
    assert "internal_location" in item and "$ref" in item["internal_location"]

    item = Item.get_record_by_pid("itemid-1")
    item = item.replace_refs()
    assert "document" in item and item["document"]["title"]
    assert "internal_location" in item and item["internal_location"]["name"]


def test_item_update(app, db, testdata):
    """Test update of an item."""
    item = Item.get_record_by_pid("itemid-1")
    item.update(dict(barcode="cm-348048"))
    item.commit()
    db.session.commit()

    assert Item.get_record_by_pid("itemid-1")["barcode"] == "CM-348048"


def test_item_validation(db, testdata):
    """Test validation when updating an item."""
    item_pid = testdata["items"][0]["pid"]
    orginal_item = Item.get_record_by_pid(item_pid)

    # change document pid
    item = deepcopy(orginal_item)
    item["document_pid"] = "not_found_doc"
    with pytest.raises(DocumentNotFoundError):
        item.commit()

    # change internal location pid
    item = deepcopy(orginal_item)
    item["internal_location_pid"] = "not_found_pid"
    with pytest.raises(InternalLocationNotFoundError):
        item.commit()


def test_item_references(db, testdata):
    """Test references when updating an item."""

    def get_active_loan_pid_and_item_pid():
        for t in testdata["items"]:
            if t["status"] == "CAN_CIRCULATE":
                item_pid = dict(type=ITEM_PID_TYPE, value=t["pid"])
                active_loan = get_active_loan_by_item_pid(item_pid).execute().hits
                total = active_loan.total.value
                if total > 0:
                    return t["pid"], active_loan[0]["pid"]

    # change document pid while is on loan
    item_pid, loan_pid = get_active_loan_pid_and_item_pid()
    item = Item.get_record_by_pid(item_pid)
    item["document_pid"] = "docid-1"
    with pytest.raises(ItemHasPastLoansError):
        item.commit()

    # change document to one that does not exist
    item = Item.get_record_by_pid("itemid-1")
    item["document_pid"] = "not_found_doc"
    with pytest.raises(DocumentNotFoundError):
        item.commit()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test records relations sequence."""

import json

from flask import url_for

from tests.helpers import get_test_record, user_login

from .helpers import (
    recrel_assert_record_relations,
    recrel_choose_endpoints_and_do_request,
)


def _test_sequence_invalid_relations_should_fail(
    client, json_headers, invalids, status_code=400
):
    """Test relation creation with invalid link relation should fail."""
    api_endpoint_series = "invenio_app_ils_relations.serid_relations"

    for invalid in invalids:
        next_pid_value = invalid["next_pid_value"]
        next_pid_type = invalid["next_pid_type"]
        previous_pid_value = invalid["previous_pid_value"]
        previous_pid_type = invalid["previous_pid_type"]
        relation_type = invalid["relation_type"]

        url = url_for(api_endpoint_series, pid_value=next_pid_value)
        payload = {
            "next_pid_value": next_pid_value,
            "next_pid_type": next_pid_type,
            "previous_pid_value": previous_pid_value,
            "previous_pid_type": previous_pid_type,
            "relation_type": relation_type,
        }

        res = client.post(url, headers=json_headers, data=json.dumps(payload))
        assert res.status_code == status_code
        if status_code == 400:
            error = json.loads(res.data.decode("utf-8"))
            assert "message" in error
            assert next_pid_value in error["message"]
            assert previous_pid_value in error["message"]


def _test_sequence(client, json_headers):
    """Test create and delete next and previous relations."""
    next_pid_value = "serid-2"
    next_pid_type = "serid"
    previous_pid_value = "serid-1"
    previous_pid_type = "serid"
    relation_type = "sequence"

    payload = {
        "next_pid_value": next_pid_value,
        "next_pid_type": next_pid_type,
        "previous_pid_value": previous_pid_value,
        "previous_pid_type": previous_pid_type,
        "relation_type": relation_type,
    }

    def _test_create_sequence(create_using_pid1=True):
        """Test relation creation of sequence Series."""

        next_rec, prev_rec = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                next_pid_value,
                next_pid_type,
                previous_pid_value,
                previous_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(
            next_rec,
            expected={
                "relations": {
                    "sequence": [
                        {
                            "pid_value": previous_pid_value,
                            "pid_type": previous_pid_type,
                            "relation_type": "sequence",
                            "relation_order": "continues",
                            "record_metadata": {
                                "title": prev_rec["title"],
                                "mode_of_issuance": prev_rec["mode_of_issuance"],
                            },
                        }
                    ]
                }
            },
        )

        recrel_assert_record_relations(
            prev_rec,
            expected={
                "relations": {
                    "sequence": [
                        {
                            "pid_value": next_pid_value,
                            "pid_type": next_pid_type,
                            "relation_type": "sequence",
                            "relation_order": "is_continued_by",
                            "record_metadata": {
                                "title": next_rec["title"],
                                "mode_of_issuance": next_rec["mode_of_issuance"],
                            },
                        }
                    ]
                }
            },
        )

    def _test_delete_sequence(create_using_pid1=True):
        """Test deletion of sequence relation."""

        next_rec, prev_rec = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                next_pid_value,
                next_pid_type,
                previous_pid_value,
                previous_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(next_rec, expected={"relations": {}})
        recrel_assert_record_relations(prev_rec, expected={"relations": {}})

    _test_create_sequence()
    _test_delete_sequence()
    _test_create_sequence(create_using_pid1=False)
    _test_delete_sequence(create_using_pid1=False)
    # recreate for the next one, to have some more valuable test data
    _test_create_sequence()


def _test_split_sequence(client, json_headers, testdata):
    """Test create split next relations."""
    next_pid_type = "serid"
    previous_pid_value = "serid-2"
    previous_pid_type = "serid"
    relation_type = "sequence"

    payload = {
        "next_pid_type": next_pid_type,
        "previous_pid_value": previous_pid_value,
        "previous_pid_type": previous_pid_type,
        "relation_type": relation_type,
    }

    # serid-1 from previous test
    rec_serid_1 = get_test_record(testdata, "series", "serid-1")
    serid_1 = {
        "pid_value": "serid-1",
        "pid_type": "serid",
        "relation_type": "sequence",
        "relation_order": "continues",
        "record_metadata": {
            "title": rec_serid_1["title"],
            "mode_of_issuance": rec_serid_1["mode_of_issuance"],
        },
    }

    def _test_create_split_sequence(create_using_pid1=True):
        """Test relation creation for split sequence."""
        pid_values = ["serid-3", "serid-4"]
        second_records = []

        for pid_value in pid_values:
            next_pid_value = pid_value
            payload["next_pid_value"] = pid_value
            next_rec, prev_rec = recrel_choose_endpoints_and_do_request(
                (client, json_headers, "POST"),
                (
                    pid_value,
                    next_pid_type,
                    previous_pid_value,
                    previous_pid_type,
                ),
                payload,
                create_using_pid1=create_using_pid1,
            )
            second_records.append(
                {
                    "pid_value": next_pid_value,
                    "pid_type": next_pid_type,
                    "relation_type": "sequence",
                    "relation_order": "is_continued_by",
                    "record_metadata": {
                        "title": next_rec["title"],
                        "mode_of_issuance": next_rec["mode_of_issuance"],
                    },
                },
            )

            # serid-2
            recrel_assert_record_relations(
                prev_rec,
                expected={"relations": {"sequence": second_records + [serid_1]}},
            )

            # serid-3 and serid-4
            recrel_assert_record_relations(
                next_rec,
                expected={
                    "relations": {
                        "sequence": [
                            {
                                "pid_value": previous_pid_value,
                                "pid_type": previous_pid_type,
                                "relation_type": "sequence",
                                "relation_order": "continues",
                                "record_metadata": {
                                    "title": prev_rec["title"],
                                    "mode_of_issuance": prev_rec["mode_of_issuance"],
                                },
                            }
                        ]
                    }
                },
            )

    _test_create_split_sequence()


def test_sequence_relations(client, json_headers, testdata, users):
    """Test sequence relations."""

    _test_sequence_invalid_relations_should_fail(
        client,
        json_headers,
        [
            {
                "next_pid_value": "serid-1",
                "next_pid_type": "serid",
                "previous_pid_value": "serid-1",
                "previous_pid_type": "serid",
                "relation_type": "sequence",
            }
        ],
        status_code=401,
    )

    user_login(client, "librarian", users)

    # serid-1 -> serid-2 (serid-1 is the previous of serid-2)
    _test_sequence(client, json_headers)

    #                                     serid-4
    # serid-2 (serid-2 is the previous) ->
    #                                     serid-3
    _test_split_sequence(client, json_headers, testdata)

    invalids = [
        {
            "next_pid_value": "serid-1",
            "next_pid_type": "serid",
            "previous_pid_value": "serid-1",
            "previous_pid_type": "serid",
            "relation_type": "sequence",
        },
        {
            "next_pid_value": "serid-1",
            "next_pid_type": "serid",
            "previous_pid_value": "docid-1",
            "previous_pid_type": "docid",
            "relation_type": "sequence",
        },
        # enable this test when the endpoint will be changed to a generic
        # relations endpoint and not for series only
        # {
        #     "next_pid_value": "docid-1",
        #     "next_pid_type": "docid",
        #     "previous_pid_value": "docid-2",
        #     "previous_pid_type": "docid",
        #     "relation_type": "sequence",
        # }
    ]
    _test_sequence_invalid_relations_should_fail(
        client, json_headers, invalids, status_code=400
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS location closures API."""

from datetime import timedelta

import arrow

from invenio_app_ils.errors import IlsException
from invenio_app_ils.proxies import current_app_ils

_ONE_DAY_INCREMENT = timedelta(days=1)  # Atomic increment


def _is_in_interval(date, interval):
    """Checks if a date is included in an interval."""
    return (
        arrow.get(interval["start_date"]).date()
        <= date
        <= arrow.get(interval["end_date"]).date()
    )


def _is_normally_open(location, date):
    """Checks if the location is normally opened on a given date."""
    opening = location["opening_weekdays"]
    opening_weekday = opening[date.weekday()]
    return opening_weekday["is_open"]


def _is_exceptionally_open(location, date):
    """Checks if the location is exceptionally opened on a given date."""
    for exception in location.get("opening_exceptions", []):
        if _is_in_interval(date, exception):
            return exception["is_open"]
    return None  # Date is not included in any interval


def _is_open_on(location, date):
    """Checks if the location is open or closed on a given date."""
    exceptionally_open = _is_exceptionally_open(location, date)
    if exceptionally_open is not None:
        return exceptionally_open
    return _is_normally_open(location, date)


def find_next_open_date(location_pid, date):
    """Finds the next day where this location is open."""
    location = current_app_ils.location_record_cls.get_record_by_pid(location_pid)
    _infinite_loop_guard = date + timedelta(days=365)
    while date < _infinite_loop_guard:
        if _is_open_on(location, date):
            return date
        date += _ONE_DAY_INCREMENT

    # Termination is normally guaranteed if there is at least one weekday open
    raise IlsException(
        description="Cannot find any date for which the "
        "location %s is open after the given date %s."
        "Please check opening/closures dates." % (location_pid, date.isoformat())
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Item serializers."""

from invenio_records_rest.serializers.csv import CSVSerializer
from invenio_records_rest.serializers.json import JSONSerializer

from invenio_app_ils.literature.serializers.custom_fields import field_cover_metadata
from invenio_app_ils.permissions import patron_permission

FILTER_KEYS = [
    "loan_pid",
    "patron_pid",
    "start_date",
    "end_date",
    "extension_count",
]


def filter_circulation(data):
    """Filter circulation status depending on user permissions."""
    if "circulation" in data["metadata"]:
        circulation = data["metadata"]["circulation"]
        patron_pid = circulation.get("patron_pid", None)
        if not patron_pid:
            return circulation

        allowed = patron_permission(patron_pid).can()

        if not allowed:
            for key in FILTER_KEYS:
                if key in circulation:
                    del circulation[key]

        data["metadata"]["circulation"] = circulation


class ItemCSVSerializer(CSVSerializer):
    """Serialize and filter item circulation status."""

    def transform_record(self, pid, record, links_factory=None, **kwargs):
        """Transform record into an intermediate representation."""
        item = super().transform_record(
            pid, record, links_factory=links_factory, **kwargs
        )
        filter_circulation(item)
        field_cover_metadata(item.get("metadata", {}).get("document", {}))
        return item

    def transform_search_hit(self, pid, record_hit, links_factory=None, **kwargs):
        """Transform search result hit into an intermediate representation."""
        hit = super().transform_search_hit(
            pid, record_hit, links_factory=links_factory, **kwargs
        )
        filter_circulation(hit)
        field_cover_metadata(hit.get("metadata", {}).get("document", {}))
        return hit


class ItemJSONSerializer(JSONSerializer):
    """Serialize and filter item circulation status."""

    def transform_record(self, pid, record, links_factory=None, **kwargs):
        """Transform record into an intermediate representation."""
        item = super().transform_record(
            pid, record, links_factory=links_factory, **kwargs
        )
        filter_circulation(item)
        field_cover_metadata(item.get("metadata", {}).get("document", {}))
        return item

    def transform_search_hit(self, pid, record_hit, links_factory=None, **kwargs):
        """Transform search result hit into an intermediate representation."""
        hit = super().transform_search_hit(
            pid, record_hit, links_factory=links_factory, **kwargs
        )
        filter_circulation(hit)
        field_cover_metadata(hit.get("metadata", {}).get("document", {}))
        return hit



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan patron resolver."""

from invenio_circulation.api import Loan


def test_loan_patron_resolver(app, testdata):
    """Test item resolving from loan."""
    loan_pid = testdata["loans"][0]["pid"]
    loan = Loan.get_record_by_pid(loan_pid)
    loan = loan.replace_refs()
    assert loan["patron_pid"] == loan["patron"]["id"]


def test_loan_patron_resolver_for_non_existing_patron_pid(app, testdata):
    """Test item resolving from loan."""
    AnonymousPatron = app.config["ILS_PATRON_ANONYMOUS_CLASS"]
    loan_pid = testdata["loans"][6]["pid"]
    loan = Loan.get_record_by_pid(loan_pid)
    loan = loan.replace_refs()
    assert loan["patron_pid"] == str(20)
    assert loan["patron"]["id"] == str(AnonymousPatron.id)
    assert loan["patron"]["name"] == "anonymous"


def test_loan_patron_resolver_for_anonymous_patron_pid(app, testdata):
    """Test item resolving from loan."""
    loan_pid = testdata["loans"][7]["pid"]
    loan = Loan.get_record_by_pid(loan_pid)
    loan = loan.replace_refs()
    assert loan["patron_pid"] == loan["patron"]["id"]
    assert loan["patron"]["name"] == "anonymous"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loans list permissions."""

import json

from flask import url_for

from tests.helpers import user_login, user_logout


def _search_loans(client, json_headers, **kwargs):
    """Return the loan fetched with a REST call."""
    url = url_for("invenio_records_rest.loanid_list", **kwargs)
    return client.get(url, headers=json_headers)


def test_anonymous_cannot_search_any_loan(client, json_headers, users):
    """Test that anonymous cannot search any loan in search results."""
    res = _search_loans(client, json_headers)
    assert res.status_code == 401


def test_admin_or_librarian_can_search_any_loan(client, json_headers, users, testdata):
    """Test that admins and librarians can search any loan."""
    for username in ["admin", "librarian"]:
        user_login(client, username, users)
        res = _search_loans(client, json_headers)
        assert res.status_code == 200
        hits = json.loads(res.data.decode("utf-8"))
        assert len(hits["hits"]["hits"]) == len(testdata["loans"])


def test_patrons_can_search_their_own_loans(client, json_headers, users, testdata):
    """Test that patrons can search their own loans."""

    def _validate_only_patron_loans(res, user, state):
        """Assert that result loans belong to the given user only."""
        patron_loans = [
            loan
            for loan in testdata["loans"]
            if loan["patron_pid"] == str(user.id) and loan["state"] == state
        ]

        assert res.status_code == 200
        hits = json.loads(res.data.decode("utf-8"))
        assert len(hits["hits"]["hits"]) == len(patron_loans)
        for hit in hits["hits"]["hits"]:
            assert hit["metadata"]["patron_pid"] == str(user.id)

    state = "PENDING"
    for username in ["patron1", "patron2"]:
        user = user_login(client, username, users)
        # search with no params
        res = _search_loans(client, json_headers, state=state)
        _validate_only_patron_loans(res, user, state)

        # search with params
        res = _search_loans(
            client,
            json_headers,
            state=state,  # test extra query
            q="patron_pid:{}".format(str(user.id)),
        )
        _validate_only_patron_loans(res, user, state)


def test_patrons_cannot_search_other_loans(client, json_headers, users, testdata):
    """Test that patrons cannot search for loans of other patrons."""
    user_login(client, "patron1", users)
    res = _search_loans(
        client,
        json_headers,
        state="PENDING",  # test extra query
        q="patron_pid:{}".format(str(users["patron2"].id)),
    )
    assert res.status_code == 403


def test_most_loaned_permissions(client, json_headers, users, testdata):
    """Test that only the backoffice is able to list the most loaned items."""
    url = url_for("invenio_app_ils_circulation_stats.most-loaned")
    tests = [
        ("admin", 200),
        ("librarian", 200),
        ("patron1", 403),
        ("anonymous", 401),
    ]
    for username, status in tests:
        user_login(client, username, users)
        res = client.get(url, headers=json_headers)
        assert res.status_code == status
        user_logout(client)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-App-Ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Provider exceptions."""

from invenio_app_ils.errors import IlsException


class ProviderNotFoundError(IlsException):
    """Raised when a provider cannot not be found."""

    description = "Provider PID '{}' was not found"

    def __init__(self, provider_pid, **kwargs):
        """Initialize exception."""
        self.description = self.description.format(provider_pid)
        super().__init__(description=self.description)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications APIs."""

import json

from flask import current_app
from invenio_records_rest.utils import obj_or_import_string

from invenio_app_ils.notifications.tasks import (
    log_error_notification,
    log_successful_notification,
)
from invenio_app_ils.proxies import current_app_ils


def build_common_msg_ctx(record):
    """Get common context for notification's message."""
    assert "patron_pid" in record

    Patron = current_app_ils.patron_cls
    patron = Patron.get_patron(record["patron_pid"])
    msg_ctx = dict(patron=patron)

    if "document_pid" in record:
        Document = current_app_ils.document_record_cls
        document = Document.get_record_by_pid(record["document_pid"])

        author = document["authors"][0]["full_name"]
        if len(document["authors"]) > 1:
            author += " et al."

        edition = document.get("edition", "")
        year = document.get("publication_year", "")
        if edition and year:
            edition_year = " ({0} - {1})".format(edition, year)
        elif edition:
            edition_year = " ({0})".format(edition)
        else:
            edition_year = " ({0})".format(year)

        full_title = "{title}, {author}{edition_year}".format(
            title=document["title"], author=author, edition_year=edition_year
        )

        msg_ctx["document"] = dict(
            pid=document["pid"],
            title=document["title"],
            author=author,
            edition=edition,
            publication_year=year,
            full_title=full_title,
        )

    return msg_ctx


def _get_notification_backends(**kwargs):
    func_or_path = current_app.config["ILS_NOTIFICATIONS_BACKENDS_BUILDER"]
    builder = obj_or_import_string(func_or_path)
    backends = builder(**kwargs)
    return backends


def send_notification(patrons, msg, **kwargs):
    """Send a notification using the configured backend(s)."""
    backends = _get_notification_backends(**kwargs)
    if not backends:
        return

    # create a serializable version of the msg for Celery tasks
    serializable_msg = msg.to_dict()
    serializable_msg.update(
        dict(
            is_manually_triggered=kwargs.pop("is_manually_triggered", False),
            patron_id=patrons[0].id,
        )
    )

    # log before sending
    log_msg = dict(
        name="notification",
        action="before_send",
        message_id=serializable_msg["id"],
        data=serializable_msg,
    )
    current_app.logger.info(json.dumps(log_msg, sort_keys=True))

    for send in backends:
        # send notification in a Celery task
        send.apply_async(
            args=[patrons, serializable_msg],
            kwargs=kwargs,
            link=log_successful_notification.s(serializable_msg),
            link_error=log_error_notification.s(data=serializable_msg),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS views."""

from flask import Blueprint, g, jsonify, render_template
from invenio_accounts.views.rest import UserInfoView, default_user_payload
from invenio_userprofiles import UserProfile


class UserInfoResource(UserInfoView):
    """Retrieve current user's information."""

    def get_user_roles(self):
        """Get all user roles."""
        return [need.value for need in g.identity.provides if need.method == "role"]

    def success_response(self, user):
        """Return response with current user's information."""
        from invenio_app_ils.proxies import current_app_ils

        user_payload = default_user_payload(user)
        user_payload["roles"] = self.get_user_roles()
        # fetch user profile for extra info
        user_profile = UserProfile.get_by_userid(user.id)

        loc_pid_value, _ = current_app_ils.get_default_location_pid
        user_payload.update(
            dict(
                username=user_profile.username,
                full_name=user_profile.full_name,
                location_pid=loc_pid_value,
            )
        )
        return jsonify(user_payload), 200


def create_logged_out_blueprint(app):
    """Create logged_out blueprint."""
    blueprint = Blueprint("logged_out", __name__)
    if app.config["DEBUG"]:
        blueprint.add_url_rule(
            "/logged-out",
            view_func=logged_out_view,
        )
    return blueprint


def logged_out_view():
    """Render logged_out view."""
    return render_template("logged_out.html")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS Document Request Decline loader JSON schema."""

from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from marshmallow import EXCLUDE, fields, validate

from invenio_app_ils.document_requests.api import DocumentRequest


class DocumentRequestDeclineSchemaV1(RecordMetadataSchemaJSONV1):
    """Document Request Decline schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    document_pid = fields.Str()
    decline_reason = fields.Str(
        required=True, validate=validate.OneOf(DocumentRequest.DECLINE_TYPES)
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test location closure cronjobs."""

import time

import arrow
from invenio_circulation.proxies import current_circulation

from invenio_app_ils.circulation.search import get_active_loans
from invenio_app_ils.closures.tasks import (
    clean_locations_past_closures_exceptions,
    extend_active_loans_location_closure,
)
from invenio_app_ils.proxies import current_app_ils

_LOCATION_PID_1 = "locid-1"
_LOCATION_PID_2 = "locid-2"
_LOAN_PID_1 = "loanid-6"
_LOAN_PID_2 = "loanid-5"


_OPENING_EXCEPTIONS_WITH_PAST_EXCEPTIONS = [
    {
        "title": "Past holidays",
        "is_open": False,
        "start_date": "2010-01-01",
        "end_date": "2010-01-06",
    },
    {
        "title": "Past holidays",
        "is_open": False,
        "start_date": "2013-04-05",
        "end_date": "2013-04-08",
    },
    {
        "title": "Past holiday",
        "is_open": False,
        "start_date": "2005-05-14",
        "end_date": "2005-05-16",
    },
    {
        "title": "Past holidays",
        "is_open": False,
        "start_date": "2019-02-01",
        "end_date": "2019-02-06",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2100-02-11",
        "end_date": "2100-02-12",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2100-03-01",
        "end_date": "2100-03-06",
    },
]

_OPENING_EXCEPTIONS_WITHOUT_CHANGES = [
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2110-01-01",
        "end_date": "2110-01-06",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2113-04-05",
        "end_date": "2113-04-08",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2105-05-14",
        "end_date": "2105-05-16",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2119-02-01",
        "end_date": "2119-02-06",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2100-02-11",
        "end_date": "2100-02-12",
    },
    {
        "title": "Future holidays",
        "is_open": False,
        "start_date": "2100-03-01",
        "end_date": "2100-03-06",
    },
]


def prepare_data(db, location_pid, opening_exceptions):
    """Adds some opening exceptions to a concrete location."""

    location = current_app_ils.location_record_cls
    record = location.get_record_by_pid(location_pid)

    record.update({"opening_exceptions": opening_exceptions})
    record.commit()
    db.session.commit()
    current_app_ils.location_indexer.index(record)


def prepare_loans_data(db, testdata):
    """Updates the expire date of a loan to match a closure day."""
    loans = testdata["loans"]

    def new_expiration_date(loan_test_data, date):
        loan = current_circulation.loan_record_cls.get_record_by_pid(
            loan_test_data["pid"]
        )
        loan.update(end_date=date.isoformat())
        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(loan)

    # Change expiration date to match closure
    date = arrow.get("2100-03-03").date()
    date_2 = arrow.get("2100-02-12").date()
    new_expiration_date(loans[5], date)
    new_expiration_date(loans[4], date_2)


def opening_exceptions_cleaned():
    """Prepare data that should match with the response."""
    opening_exceptions = [
        {
            "title": "Future holidays",
            "is_open": False,
            "start_date": "2100-02-11",
            "end_date": "2100-02-12",
        },
        {
            "title": "Future holidays",
            "is_open": False,
            "start_date": "2100-03-01",
            "end_date": "2100-03-06",
        },
    ]

    return opening_exceptions


def test_clean_locations_past_closures_exceptions(db, users, testdata):
    """Test cleaning of past exceptions."""
    prepare_data(db, _LOCATION_PID_1, _OPENING_EXCEPTIONS_WITH_PAST_EXCEPTIONS)
    prepare_data(db, _LOCATION_PID_2, _OPENING_EXCEPTIONS_WITH_PAST_EXCEPTIONS)
    clean_locations_past_closures_exceptions()
    location = current_app_ils.location_record_cls
    record = location.get_record_by_pid(_LOCATION_PID_1)
    record_2 = location.get_record_by_pid(_LOCATION_PID_2)

    assert record["opening_exceptions"] == opening_exceptions_cleaned()
    assert record_2["opening_exceptions"] == opening_exceptions_cleaned()


def test_no_changes_on_loans(db, users, testdata):
    """Test that no changes are applied to the loans if not needed."""
    prepare_data(db, _LOCATION_PID_1, _OPENING_EXCEPTIONS_WITHOUT_CHANGES)
    loans = testdata["loans"]
    clean_locations_past_closures_exceptions()
    for hit in get_active_loans().scan():
        for loan in loans:
            if hit["pid"] == loan["pid"]:
                assert hit["end_date"] == loan["end_date"]


def test_no_changes_on_exceptions(db, users, testdata):
    """Test that no changes are applied to the exceptions if not needed."""
    prepare_data(db, _LOCATION_PID_1, _OPENING_EXCEPTIONS_WITHOUT_CHANGES)
    extend_active_loans_location_closure(_LOCATION_PID_1)
    location = current_app_ils.location_record_cls
    record = location.get_record_by_pid(_LOCATION_PID_1)
    record_2 = location.get_record_by_pid(_LOCATION_PID_2)

    assert record["opening_exceptions"] == _OPENING_EXCEPTIONS_WITHOUT_CHANGES
    assert record_2["opening_exceptions"] == []


def test_extend_active_loans_location_closure(db, users, testdata, app_with_notifs):
    """Test extension of active loans that finish on holidays."""
    prepare_data(db, _LOCATION_PID_1, _OPENING_EXCEPTIONS_WITH_PAST_EXCEPTIONS)
    prepare_loans_data(db, testdata)
    time.sleep(3)

    with app_with_notifs.extensions["mail"].record_messages() as outbox:
        assert len(outbox) == 0
        extend_active_loans_location_closure(_LOCATION_PID_1)
        assert len(outbox) == 2

    loan = current_circulation.loan_record_cls.get_record_by_pid(_LOAN_PID_1)
    loan_2 = current_circulation.loan_record_cls.get_record_by_pid(_LOAN_PID_2)
    assert loan["end_date"] == "2100-03-07"
    assert loan_2["end_date"] == "2100-02-13"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Location closures tests."""

import json

import arrow
from flask import url_for

from invenio_app_ils.closures.api import find_next_open_date
from invenio_app_ils.proxies import current_app_ils
from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
_LOCATION_PID = "locid-1"
_LOCATION_NAME = "Location name"
_ITEM_ENDPOINT = "invenio_records_rest.locid_item"
_LIST_ENDPOINT = "invenio_records_rest.locid_list"
_WEEKDAYS = [
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]
_DEFAULT_TIMES = [
    {"start_time": "08:00", "end_time": "12:00"},
    {"start_time": "13:00", "end_time": "18:00"},
]


def _build_location_closures_data(closed_weekdays, exceptions):
    opening_weekdays = []
    for name in _WEEKDAYS:
        is_open = name not in closed_weekdays
        obj = {"weekday": name, "is_open": is_open}
        if is_open:
            obj["times"] = _DEFAULT_TIMES
        opening_weekdays.append(obj)

    opening_exceptions = []
    for start_date, end_date, is_open in exceptions:
        opening_exceptions.append(
            {
                "title": "%s - %s" % (start_date, end_date),
                "start_date": start_date,
                "end_date": end_date,
                "is_open": is_open,
            }
        )

    return {
        "opening_weekdays": opening_weekdays,
        "opening_exceptions": opening_exceptions,
    }


def _date_from_string(date_string):
    return arrow.get(date_string).date()


def test_location_permissions(client, testdata, json_headers, users):
    """Test location endpoints permissions."""
    dummy_location = dict(
        name=_LOCATION_NAME,
        opening_weekdays=[
            {"weekday": w, "is_open": True, "times": _DEFAULT_TIMES} for w in _WEEKDAYS
        ],
        opening_exceptions=[],
    )
    tests = [
        ("admin", _HTTP_OK, dummy_location),
        ("librarian", _HTTP_OK, dummy_location),
        ("patron1", [403], dummy_location),
        ("anonymous", [401], dummy_location),
    ]
    read_statuses = [200]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(_LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_create(expected_status, data):
        """Test record creation."""
        url = url_for(_LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == _LOCATION_NAME
            return record["pid"]

    def _test_update(expected_status, data, pid):
        """Test record update."""
        pid_value = pid or _LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == _LOCATION_NAME

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or _LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        pid_value = pid or _LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status, data in tests:
        user_login(client, username, users)
        _test_list(read_statuses)
        pid = _test_create(expected_status, data)
        _test_update(expected_status, data, pid)
        _test_read(read_statuses, pid)
        _test_delete(expected_status, pid)


def test_location_validation(client, json_headers, users, testdata):
    def _test_update_location_closures(data, expected_code):
        url = url_for("invenio_records_rest.locid_item", pid_value=_LOCATION_PID)
        res = client.get(url, headers=json_headers)
        assert res.status_code == 200
        metadata = res.get_json()["metadata"]
        metadata.update(data)
        res = client.put(url, headers=json_headers, data=json.dumps(metadata))
        assert res.status_code == expected_code

    def _test_update_weekdays(weekdays, expected_code):
        data = [
            {
                "weekday": name,
                "is_open": is_open,
                **({"times": _DEFAULT_TIMES} if is_open else {}),
            }
            for name, is_open in weekdays
        ]
        _test_update_location_closures(
            {"opening_weekdays": data, "opening_exceptions": []}, expected_code
        )

    def _test_update_times(times, expected_code):
        ref = "monday"
        times_data = [{"start_time": start, "end_time": end} for start, end in times]
        data = [
            {
                "weekday": w,
                "is_open": w == ref,
                **({"times": times_data} if w == ref else {}),
            }
            for w in _WEEKDAYS
        ]
        _test_update_location_closures(
            {"opening_weekdays": data, "opening_exceptions": []}, expected_code
        )

    def _test_update_exceptions(closed_weekdays, exceptions, expected_code):
        _test_update_location_closures(
            _build_location_closures_data(closed_weekdays, exceptions),
            expected_code,
        )

    user_login(client, "librarian", users)

    # Weekdays

    _test_update_weekdays([[w, True] for w in _WEEKDAYS], 200)
    _test_update_weekdays([[w, True] for w in _WEEKDAYS[::-1]], 200)
    _test_update_weekdays([[w, False] for w in _WEEKDAYS], 400)
    _test_update_weekdays([[w, True] for w in _WEEKDAYS[:6]], 400)
    _test_update_weekdays([[w, True] for w in _WEEKDAYS[:6] + ["monday"]], 400)
    _test_update_weekdays([["foobar", True]], 400)

    # Hours

    _test_update_times([["08:00", "12:00"], ["13:00", "18:00"]], 200)
    _test_update_times([["13:00", "18:00"], ["08:00", "12:00"]], 200)
    _test_update_times([["08:00", "18:00"]], 200)
    _test_update_times([], 400)
    _test_update_times(
        [["08:00", "10:00"], ["11:00", "12:00"], ["13:00", "18:00"]], 400
    )
    _test_update_times([["8:00", "12:00"], ["13:00", "18:00"]], 400)
    _test_update_times([["08:00", "12:0"], ["13:00", "18:00"]], 400)
    _test_update_times([["08:00", ""], ["13:00", "18:00"]], 400)
    _test_update_times([["08:00", "12:00"], ["12:00", "18:00"]], 400)
    _test_update_times([["08:00", "12:00"], ["09:00", "11:00"]], 400)

    # Exceptions

    _test_update_exceptions(
        ["saturday", "sunday"],
        [
            ["2000-01-01", "2000-01-05", False],
            ["2000-01-07", "2000-01-09", True],
            ["2000-01-10", "2000-01-15", True],
        ],
        200,
    )

    _test_update_exceptions(
        ["saturday", "sunday"],
        [
            ["2000-01-12", "2000-01-17", True],
            ["2000-01-07", "2000-01-11", False],
            ["2000-01-02", "2000-01-04", True],
        ],
        200,
    )

    _test_update_exceptions(
        [],
        [
            ["2000-01-01", "2000-01-05", False],
            ["2000-01-04", "2000-01-08", False],
        ],
        400,
    )

    _test_update_exceptions(
        [],
        [
            ["2000-01-01", "2000-01-05", False],
            ["2000-01-04", "2000-01-08", True],
        ],
        400,
    )

    _test_update_exceptions(
        [],
        [
            ["2000-01-01", "2000-01-01", False],
            ["2000-01-01", "2000-01-01", False],
        ],
        400,
    )

    _test_update_exceptions(
        [],
        [
            ["2000-01-02", "2000-01-01", True],
        ],
        400,
    )


def test_find_next_open_date(app, db, testdata):
    def _update_location_closures_data(closed_weekdays, exceptions):
        Location = current_app_ils.location_record_cls
        record = Location.get_record_by_pid(_LOCATION_PID)

        record.update(_build_location_closures_data(closed_weekdays, exceptions))
        record.commit()
        db.session.commit()
        current_app_ils.location_indexer.index(record)

        return record

    def _test(start_date, expected_next_open_date):
        next_open = find_next_open_date(_LOCATION_PID, _date_from_string(start_date))
        if expected_next_open_date:
            assert next_open == _date_from_string(expected_next_open_date)
        else:
            assert next_open is None

    """
    Mon. Tue. Wed. Thu. Fri. Sat. Sun.
     27   28   29   30   31  -01  -02
    x03  x04  x05   06   07  -08  -09
    x10  x11  x12  x13  x14  x15  x16
    x17  x18  x19  x20  x21  x22  -23
    x24   25   26  x27  x28  o29  o30
     31   01   02   03   04  -05  -06
    """

    closed_weekdays = ["saturday", "sunday"]
    exceptions = [
        ["2000-01-03", "2000-01-05", False],  # Mon. - Wed.
        ["2000-01-09", "2000-01-22", False],  # Sun. - Sat.
        ["2000-01-24", "2000-01-24", False],  # Mon.
        ["2000-01-27", "2000-01-28", False],  # Thu. - Fri.
        ["2000-01-29", "2000-01-30", True],  # Sat. - Sun.
    ]
    _update_location_closures_data(closed_weekdays, exceptions)

    _test("2000-01-01", "2000-01-06")
    _test("2000-01-04", "2000-01-06")
    _test("2000-01-06", "2000-01-06")
    _test("2000-01-07", "2000-01-07")

    _test("2000-01-09", "2000-01-25")
    _test("2000-01-13", "2000-01-25")

    _test("2000-01-26", "2000-01-26")
    _test("2000-01-27", "2000-01-29")
    _test("2000-01-30", "2000-01-30")

    _test("2000-02-05", "2000-02-07")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests documents CRUD."""

from invenio_app_ils.documents.api import Document


def _assert_extra_fields(doc):
    """Test that extra fields are automatically added."""
    assert "$schema" in doc
    assert "circulation" in doc and "$ref" in doc["circulation"]
    assert "relations" in doc and "$ref" in doc["relations"]
    assert "eitems" in doc and "$ref" in doc["eitems"]
    assert "items" in doc and "$ref" in doc["items"]
    assert "stock" in doc and "$ref" in doc["stock"]


def test_document_creation_refs(app):
    """Test creation of a document."""
    d = dict(
        pid="a1bc",
        created_by={"type": "script", "value": "demo"},
        title="Test title",
        authors=[dict(full_name="John Doe")],
        publication_year="2010",
        document_type="BOOK",
    )
    doc = Document.create(d)

    _assert_extra_fields(doc)


def test_document_update_refs(app):
    """Test update of a document."""
    d = dict(
        pid="a1bc",
        title="Test title",
        created_by={"type": "script", "value": "demo"},
        authors=[dict(full_name="John Doe")],
        publication_year="2010",
        document_type="BOOK",
    )
    doc = Document.create(d)
    del doc["circulation"]
    del doc["relations"]
    del doc["eitems"]
    del doc["items"]
    del doc["stock"]
    doc.update(dict(title="Test title 2"))

    _assert_extra_fields(doc)
    assert doc["title"] == "Test title 2"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test document requests."""

import json

from flask import url_for

from tests.helpers import user_login


def _compare_list_data(all_expected, data, user_id, is_admin):
    """Compare list data with expected values."""
    if is_admin:
        expected = []
        for pids in all_expected.values():
            for pid in pids:
                expected.append(pid)
    else:
        expected = all_expected[str(user_id)]

    for hit in data["hits"]["hits"]:
        assert hit["metadata"]["pid"] in expected


def test_request_list_permissions(client, testdata, json_headers, users):
    """Test doc req list permissions."""
    url = url_for("invenio_records_rest.dreqid_list")
    res = client.get(url, headers=json_headers)
    assert res.status_code == 401

    requests_by_patron = {}
    for req in testdata["document_requests"]:
        if req["patron_pid"] not in requests_by_patron:
            requests_by_patron[req["patron_pid"]] = []
        requests_by_patron[req["patron_pid"]].append(req["pid"])

    for username in users.keys():
        user = user_login(client, username, users)
        url = url_for("invenio_records_rest.dreqid_list")
        res = client.get(url, headers=json_headers)
        assert res.status_code == 200
        data = json.loads(res.data.decode("utf-8"))
        is_admin = "librarian" in user.email or "admin" in user.email
        _compare_list_data(requests_by_patron, data, user.id, is_admin)


def test_create_document_request(client, testdata, json_headers, users):
    """Test creating document requests."""
    tests = [
        (
            "admin",
            201,
            dict(
                title="Test title",
                request_type="LOAN",
                medium="Paper",
                patron_pid="1",
            ),
        ),
        (
            "librarian",
            400,
            dict(
                patron_pid="1",
            ),
        ),
        (
            "admin",
            201,
            dict(
                patron_pid="1",
                title="Test",
                invalid_param="Test",
                request_type="LOAN",
                medium="Paper",
            ),
        ),
        (
            "librarian",
            201,
            dict(
                title="Test title",
                patron_pid="1",
                request_type="BUY",
                medium="Paper",
            ),
        ),
        (
            "patron1",
            201,
            dict(
                title="Test title",
                patron_pid="1",
                request_type="BUY",
                medium="Paper",
            ),
        ),
        (
            "patron2",
            400,
            dict(
                title="Test title",
                patron_pid="1",
                request_type="LOAN",
                medium="Paper",
            ),
        ),
        (
            "patron2",
            201,
            dict(
                title="Test title",
                patron_pid="2",
                request_type="LOAN",
                medium="Paper",
            ),
        ),
    ]
    for username, expected_status, data in tests:
        user_login(client, username, users)
        url = url_for("invenio_records_rest.dreqid_list")
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code == expected_status
        if res.status_code == 201:
            assert "invalid_param" not in json.loads(res.data)["metadata"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Metadata Extensions."""

from copy import deepcopy

from flask import current_app
from invenio_records_rest.schemas.fields import DateString, SanitizedUnicode
from marshmallow import Schema
from marshmallow.fields import Bool, Integer, List


class MetadataExtensions(object):
    """Custom metadata extensions helper class."""

    def __init__(self, namespaces, extensions):
        """Constructor."""
        self.namespaces = deepcopy(namespaces) or {}
        self.extensions = deepcopy(extensions) or {}
        self._validate()

    def _validate_marshmallow_type(self, field_cfg):
        """Make sure the Marshmallow type is one we support."""

        def validate_basic_marshmallow_type(_type):
            allowed_types = [Bool, DateString, Integer, SanitizedUnicode]
            assert any(
                [isinstance(_type, allowed_type) for allowed_type in allowed_types]
            )

        marshmallow_type = field_cfg["marshmallow"]
        if isinstance(marshmallow_type, List):
            validate_basic_marshmallow_type(marshmallow_type.inner)
        else:
            validate_basic_marshmallow_type(marshmallow_type)

    def _validate_elasticsearch_type(self, field_cfg):
        """Make sure the Elasticsearch type is one we support."""
        allowed_types = ["boolean", "date", "long", "keyword", "text"]
        assert field_cfg["elasticsearch"] in allowed_types

    def _validate(self):
        """Validates extension configuration.

        We only allow certain types, so this private method flags divergence
        from what is allowed early.
        """
        # Validate namespaces
        contexts = set()
        for settings in self.namespaces.values():
            assert settings["@context"] not in contexts
            contexts.add(settings["@context"])

        # Validate extensions
        for field_key, field_cfg in self.extensions.items():
            self._validate_marshmallow_type(field_cfg)
            self._validate_elasticsearch_type(field_cfg)

    def to_schema(self):
        """Dynamically creates and returns the extensions Schema."""
        schema_dict = {
            field_key: field_cfg["marshmallow"]
            for field_key, field_cfg in self.extensions.items()
        }
        return Schema.from_dict(schema_dict)

    def get_field_type(self, field_key, _type):
        """Returns type value for given field_key and _type.

        :params field_key: str formatted as <prefix>:<field_id>
        :params _type: str "elasticsearch" or "marshmallow"
        """
        return self.extensions.get(field_key, {}).get(_type)


def add_es_metadata_extensions(record_dict):
    """Add "extensions_X" fields to record_dict prior to Elasticsearch index.

    :param record_dict: dumped Record dict
    """
    rec_type = record_dict["$schema"].split("/")[-1].split("-")[0]
    metadata_extensions = getattr(
        current_app.extensions["invenio-app-ils"],
        "{}_metadata_extensions".format(rec_type),
    )

    for key, value in record_dict.get("extensions", {}).items():
        field_type = metadata_extensions.get_field_type(key, "elasticsearch")
        if not field_type:
            continue

        es_field = "extensions_{}s".format(field_type)

        if es_field not in record_dict:
            record_dict[es_field] = []

        record_dict[es_field].append({"key": key, "value": value})
